import numpy as np
import math
from santex import Tensor

import matplotlib.pyplot as plt
import numpy as np
from .vtkplotter import Plotter
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from .plot_vel_grid import plot_velocity_grid
from .utils import christoffel_tensor, wave_property

from .. import Material
from .. import EBSD

from scipy.interpolate import griddata
import math

class Anisotropy:
    """
    Anisotropy class for representing anisotropic material properties.

    Attributes:
        cijkl (numpy.ndarray or None): stiffness tensor in voigt notation.
            It is converted from the input stiffness matrix using voigt_to_tensor method.
            If stiffness_matrix is None, cijkl remains None. Units can be SI or CGS but should remain consistent
            between stiffness matrix and density
        rho (float or None): Density of the material.
            If density is None, rho remains None.

    Methods:
        __init__(stiffness_matrix, density):
            Initializes the Anisotropy object with the given stiffness_matrix and density.
            If either stiffness_matrix or density is None, cijkl and rho remain None.

    Example Usage:
        stiffness_matrix = np.array([[198.96, 73.595, 68.185, 0., 9.735, 0.],
                                [73.595, 155.94, 62.23, 0., 6.295, 0.],
                                [68.185, 62.23, 225.99, 0., 33.85, 0.],
                                [0., 0., 0., 65.66, 0., 6.415],
                                [9.735, 6.295, 33.85, 0., 60.23, 0.],
                                [0., 0., 0., 6.415, 0., 65.18]]) * 10**9
        density = 3.5 
        anisotropy = Anisotropy(stiffness_matrix, density)
    """
    
    def __init__(self, stiffness_matrix, density):
        """
        Initializes an Anisotropy object with the given stiffness_matrix and density.

        Parameters:
            stiffness_matrix (list of lists or numpy.ndarray): The 6x6 stiffness matrix representing
                the material's anisotropic properties in Voigt notation.
            density (float): The density of the material.

        Raises:
            TypeError: If stiffness_matrix is not a 6x6 list or numpy.ndarray.

        """
        self.cijkl = None
        self.rho = None

        if stiffness_matrix is not None and density is not None:
            tensor_object = Tensor()
            self.cijkl = tensor_object.voigt_to_tensor(stiffness_matrix)
            self.rho = density


    def christoffel_tensor(self, n):
        """
        Calculates the Christoffel tensor given a direction vector n.

        Parameters:
            n (numpy.ndarray): The direction vector for which the Christoffel tensor is calculated.

        Returns:
            numpy.ndarray: The Christoffel tensor Tik for the given direction vector n.

        Raises:
            ValueError: If an error occurs during the calculation.

        Example direction:
            n = np.array([1, 0, 0])
        
        """
        try:
            return christoffel_tensor(self.cijkl, n)
        except Exception as e:
            raise ValueError("Error in calculating the Christoffel tensor:", e)

    def wave_property(self, tik):
        """
        Calculates the wave properties (wave moduli and polarization directions) given the Christoffel tensor Tik.

        Parameters:
            tik (numpy.ndarray): The Christoffel tensor Tik for which wave properties are calculated.

        Returns:
            tuple: A tuple containing the wave moduli and polarization directions.

        Raises:
            ValueError: If an error occurs during the calculation.
        """
        try:
            return wave_property(tik)
        except Exception as e:
            raise ValueError("Error in calculating wave properties:", e)

    def phase_velocity(self):
        """
        Calculates the phase velocities (P-wave velocity, S1-wave velocity, and S2-wave velocity) for different
        directions using the Christoffel tensor and wave properties.

        Returns:
            tuple: A tuple containing the phase velocities (vp for P-wave, vs1 for S1-wave, and vs2 for S2-wave).

        Raises:
            ValueError: If an error occurs during the calculation.    
        """
        try:
            vp = []
            vs1 = []
            vs2 = []

            theta_values = np.arange(0, math.pi + math.pi / 180, math.pi / 180)
            phi_values = np.arange(0, 2 * math.pi + math.pi / 180, math.pi / 180)

            theta, phi = np.meshgrid(theta_values, phi_values, indexing='ij')

            sin_theta = np.sin(theta)
            cos_theta = np.cos(theta)
            sin_phi = np.sin(phi)
            cos_phi = np.cos(phi)

            sin_theta_sin_phi = sin_theta * cos_phi
            sin_theta_cos_phi = sin_theta * sin_phi

            sqrt_rho = np.sqrt(self.rho)

            for i in range(theta.shape[0]):
                for j in range(phi.shape[1]):
                    n = np.array([sin_theta_sin_phi[i, j], sin_theta_cos_phi[i, j], cos_theta[i, j]])
                    tik = self.christoffel_tensor(n)
                    if tik is not None:
                        wave_moduli, polarization_directions = self.wave_property(tik)
                        vp.append(np.sqrt(wave_moduli[0] / self.rho))
                        vs1.append(np.sqrt(wave_moduli[1] / self.rho))
                        vs2.append(np.sqrt(wave_moduli[2] / self.rho))

            return tuple(vp), tuple(vs1), tuple(vs2)
        except Exception as e:
            raise ValueError("Error in calculating phase velocity:", e)
        
    def velocities(self):
        vp, vs1, vs2 = self.phase_velocity()
        return vp, vs1, vs2

    def anisotropy_values(self, stiffness_matrix = None, density = None, method = None, return_values=None):
        """
        Calculates various anisotropy values based on the velocities calculated from the given stiffness matrix and density.

        Parameters:
            stiffness_matrix (list or None): The stiffness matrix representing the material's anisotropic properties.
            density (float or None): The density of the material.
            method (str or None): The method to use for calculating anisotropy values. Options: 'array'.
            return_values (str or None): The specific anisotropy value to return. Options: 'maxvp', 'minvp', 'maxvs1',
                'minvs1', 'maxvs2', 'minvs2', or None (default) to print all values.

        Returns:
            float or None: A dictionary containing the calculated anisotropy values if return_values is None,
                or a single anisotropy value if return_values is specified.

        Raises:
            ValueError: If an error occurs during the calculation.

        Notes:
            - If method='array', an array of Anisotropy objects can be provided for batch calculation.
        
        """

        if method == "array":
            anis_mat = []
            for i in stiffness_matrix:
                anis = Anisotropy(stiffness_matrix[i], density[i])
                anis_mat.append(self.anisotropy_values(anis))
                self.anisotropy_values(anis)
            return anis_mat

        vp, vs1, vs2 = self.velocities()
        maxvp = max(vp)
        minvp = min(vp)
        maxvs1 = max(vs1)
        minvs1 = min(vs1)
        maxvs2 = max(vs2)
        minvs2 = min(vs2)
        swaveAnisotropy_percent = 200 * (np.array(vs1) - np.array(vs2)) / (np.array(vs1) + np.array(vs2))
        max_vs_anisotropy_percent = max(swaveAnisotropy_percent)
        min_vs_anisotropy_percent = min(swaveAnisotropy_percent)
        p_wave_anisotropy_percent = 200 * (maxvp - minvp) / (maxvp + minvp)
        s1_wave_anisotropy_percent = 200 * (maxvs1 - minvs1) / (maxvs1 + minvs1)
        s2_wave_anisotropy_percent = 200 * (maxvs2 - minvs2) / (maxvs2 + minvs2)
        dvs = np.array(vs1) - np.array(vs2)
        maxdvs = max(dvs)
        vp_vs1 = np.array(vp) / np.array(vs1)
        AVpVs1 = 200 * (max(vp_vs1) - min(vp_vs1)) / (max(vp_vs1) + min(vp_vs1))

        if return_values == 'maxvp':
            return maxvp
        elif return_values == 'maxvs1':
            return maxvs1
        elif return_values == 'maxvs1':
            return maxvs1
        elif return_values == 'maxvs1':
            return maxvs1
        elif return_values == 'maxvs1':
            return maxvs1
        elif return_values == 'maxvs1':
            return maxvs1
        elif return_values == 'maxvs1':
            return maxvs1
        elif return_values == 'maxvs1':
            return maxvs1
        
        else:
            print("Max Vp: ", maxvp)
            print("Min Vp: ", minvp)
            print("Max Vs1: ", maxvs1)
            print("Min Vs1: ", minvs1)
            print("Max Vs2: ", maxvs2)
            print("Min Vs2: ", minvs2)
            print("Max vs anisotropy percent: ", max_vs_anisotropy_percent)
            print("Min vs anisotropy percent: ", min_vs_anisotropy_percent)
            print("P wave anisotropy percent: ", p_wave_anisotropy_percent)
            print("S1 Wave anisotropy percent: ", s1_wave_anisotropy_percent)
            print("S2 Wave anisotropy percent: ", s2_wave_anisotropy_percent)
            print("Velocity difference: ", maxdvs)
            print("Vp/Vs1 ratio: ", AVpVs1)
            return {
                'maxvp': maxvp,
                'minvp': minvp,
                'maxvs1': maxvs1,
                'minvs1': minvs1,
                'maxvs2': maxvs2,
                'minvs2': minvs2,
                'max_vs_anisotropy_percent': max_vs_anisotropy_percent,
                'min_vs_anisotropy_percent': min_vs_anisotropy_percent,
                'p_wave_anisotropy_percent': p_wave_anisotropy_percent,
                's1_wave_anisotropy_percent': s1_wave_anisotropy_percent,
                's2_wave_anisotropy_percent': s2_wave_anisotropy_percent,
                'maxdvs': maxdvs,
                'AVpVs1': AVpVs1
            }
        
    def plot_velocities(self, pressure_range, temperature_range, return_type, is_ebsd = False, phase = None, grid = [5, 5], filename = None, *args):
        """
        Plots velocities based on specified ranges and return types.

            Parameters:
                pressure_range (tuple): The range of pressures for which velocities will be plotted. Example: [2, 4].
                temperature_range (tuple): The range of temperatures for which velocities will be plotted. Example: [1000, 2000]
                return_type (str): The type of velocity to plot. Options: 'maxvp', 'minvp', 'maxvs1', 'minvs1', 'maxvs2', 'minvs2'.
                is_ebsd (bool): Whether the data comes from electron backscatter diffraction (EBSD). Default is False.
                phase (str or None): The phase of the material. Example, phase = 'Forsterite'
                grid (list): The grid dimensions for the plot. Default is [5, 5].
                filename (str or None): The filename of the EBSD data. Required if is_ebsd is True.
                *args: can be [1, 2, 3]

            Raises:
                ValueError: If required parameters are not provided or if an error occurs during plotting.

            Notes:
                - If is_ebsd is True, phase and filename must be provided.
                - The plot type and appearance can be customized using *args.

            Example Usage:
                anisotropy.plot_velocities((0, 100), (500, 1000), 'maxvp', is_ebsd=True, phase='phase1', filename='velocity_plot.png', 'ro-')
        """
        return plot_velocity_grid(pressure_range=pressure_range, temperature_range=temperature_range, return_type=return_type, is_ebsd=is_ebsd, phase = phase, grid = grid, filename = filename, *args)



    def plot(self, colormap="RdBu_r", step = 180, savefig = False, figname = None, dpi = 300):
        """
        Plots various anisotropic maps based on the Christoffel tensor.

        Parameters:
            colormap (str): The colormap to use for plotting. Default is "RdBu_r".
            step (int): The step size for theta and phi values. Default is 180.
            savefig (bool): Whether to save the plot as an image. Default is False.
            figname (str or None): The filename to save the plot. Required if savefig is True.
            dpi (int): The resolution of the saved image. Default is 300.

        Raises:
            ValueError: If an error occurs during the plotting process.

        Notes:
            - This method generates a 2x3 grid of subplots, each representing different anisotropic maps based on the Christoffel tensor.
            - The colormap, step size, and other parameters can be customized.

        Example Usage:
            anisotropy.plot(colormap="viridis", step=120, savefig=True, figname="anisotropy_plot.png", dpi=600)
        """
        try:
            fig, axs = plt.subplots(2, 3, figsize=(15, 10))
            step = math.pi / step

            # texts for each subplot
            texts = ['Ratio of VP to VS1', 'Velocity of P-waves (VP)', 'Velocity of S1-waves (VS1)', 
                    'Velocity of S2-waves (VS2)', 'Anisotropy measure for VP and VS1', 
                    'Anisotropy measure for VP and VS2']
            
            for i, ax in enumerate(axs.flat):
                x = []
                y = []
                c = []

                for theta in np.arange(0, math.pi / 2 + step, step):
                    for phi in np.arange(0, 2 * math.pi + step, step):
                        n = np.array([math.sin(theta) * math.cos(phi), math.sin(theta) * math.sin(phi), math.cos(theta)])
                        tik = self.christoffel_tensor(n)
                        if tik is not None:
                            wave_moduli, _ = self.wave_property(tik)
                            vp = math.sqrt(wave_moduli[0] / self.rho)
                            vs1 = math.sqrt(wave_moduli[1] / self.rho)
                            vs2 = math.sqrt(wave_moduli[2] / self.rho)
                            vpvs1 = vp/vs1
                            x.append(n[0] / (1 + n[2]))
                            y.append(n[1] / (1 + n[2]))
                            if i == 0:
                                c.append(vpvs1)
                            elif i == 1:
                                c.append(vp)
                            elif i == 2:
                                c.append(vs1)
                            elif i == 3:
                                c.append(vs2)
                            elif i == 4:
                                c.append((vp - vs1) / (vp + vs1))
                            elif i == 5:
                                c.append((vp - vs2) / (vp + vs2))

                # Interpolate onto a regular grid
                xi = np.linspace(min(x), max(x), 100)
                yi = np.linspace(min(y), max(y), 100)
                xi, yi = np.meshgrid(xi, yi)
                zi = griddata((x, y), c, (xi, yi), method='linear')

                # Plotting contour lines for each subplot
                contours = ax.contour(xi, yi, zi, 5, colors='black')
                ax.clabel(contours, inline=True, fontsize=8)

                sc = ax.scatter(x, y, c=c, cmap=colormap, s=5)  # Reduce scatter dot size s for clarity
                ax.set_xlabel('x')
                ax.set_ylabel('y')
                ax.set_aspect('equal', 'box')

                # Adding above text at the bottom of each plot
                ax.text(0.5, -0.15, texts[i], ha='center', transform=ax.transAxes)

            axs[0, 0].set_title('VP/VS1')
            axs[0, 1].set_title('VP')
            axs[0, 2].set_title('VS1')
            axs[1, 0].set_title('VS2')
            axs[1, 1].set_title('AVpVs1')
            axs[1, 2].set_title('AVpVs2')

            plt.tight_layout()
            plt.show()

            if savefig:
                plt.savefig(f"{figname}", dpi = dpi)
        except Exception as e:
            print(f"An error occurred: {e}")


    def plotly(self):
        """
        Generates a Plotly figure to visualize material properties based on calculations using Christoffel and wave tensors.

        This method creates a Plotly figure with subplots for various properties such as VP/VS1, VP, VS1, VS2, AVpVs1, and AVpVs2.
        The figure is generated by iterating over specified ranges of angles and calculating properties like wave moduli, VP, VS1, VS2,
        VP/VS1 ratios, and AVpVs1/AVpVs2 ratios. The resulting data points are plotted on scatter plots with a color scale representing
        different property values.

        Raises:
            ValueError: If an error occurs during the plotting process.

        Notes:
            - Requires the existence of the `christoffel_tensor` and `wave_property` methods in the class instance to calculate material properties.

        Example Usage:
            plotter = anisotropy_instance.plotly()
        """
        try:
            fig = go.Figure()

            step = math.pi / 180

            fig = make_subplots(rows=2, cols=3, subplot_titles=("VP/VS1", "VP", "VS1", "VS2", "AVpVs1", "AVpVs2"))

            for i in range(6):
                x = []
                y = []
                c = []

                for theta in np.arange(0, math.pi / 2 + step, step):
                    for phi in np.arange(0, 2 * math.pi + step, step):
                        n = np.array([math.sin(theta) * math.cos(phi), math.sin(theta) * math.sin(phi), math.cos(theta)])
                        tik = self.christoffel_tensor(n)
                        if tik is not None:
                            wave_moduli, _ = self.wave_property(tik)
                            vp = math.sqrt(wave_moduli[0] / self.rho)
                            vs1 = math.sqrt(wave_moduli[1] / self.rho)
                            vs2 = math.sqrt(wave_moduli[2] / self.rho)
                            vpvs1 = vp/vs1
                            x.append(n[0] / (1 + n[2]))
                            y.append(n[1] / (1 + n[2]))
                            if i == 0:
                                c.append(vpvs1)
                            elif i == 1:
                                c.append(vp)
                            elif i == 2:
                                c.append(vs1)
                            elif i == 3:
                                c.append(vs2)
                            elif i == 4:
                                c.append((vp - vs1) / (vp + vs1))
                            elif i == 5:
                                c.append((vp - vs2) / (vp + vs2))

                row = i // 3 + 1
                col = i % 3 + 1
                fig.add_trace(go.Scatter(
                    x=x,
                    y=y,
                    mode='markers',
                    marker=dict(
                        size=5,
                        color=c,
                        colorscale='RdBu',
                        colorbar=dict(title='Colorbar Title'),
                    ),
                    name='',
                ), row=row, col=col)

            fig.update_layout(
                title="Plot Title",
                xaxis_title="x",
                yaxis_title="y"
            )

            fig.show()
        except Exception as e:
            raise ValueError("Error in plotting:", e)


            
    def plotter_vs_splitting(self, density, voigt_stiffness):
        """
        Plots the Vs splitting based on density and Voigt stiffness.

        Parameters:
            density (float): The density of the material.
            voigt_stiffness (float): The Voigt stiffness of the material.
        """
        Plotter.plot_vs_splitting(voigt_stiffness, density)

    def plotter_vp(self, density, voigt_stiffness):
        """
        Plots the Vp based on density and Voigt stiffness.

        Parameters:
            density (float): The density of the material.
            voigt_stiffness (float): The Voigt stiffness of the material.
        """

        Plotter.plot_vp(voigt_stiffness, density)

    def plotter_vs1(self, density, voigt_stiffness):
        """
        Plots the Vs1 based on density and Voigt stiffness.

        Parameters:
            density (float): The density of the material.
            voigt_stiffness (float): The Voigt stiffness of the material.
        """

        Plotter.plot_vp(voigt_stiffness, density)

    def plotter_vs2(self, density, voigt_stiffness):
        """
        Plots the Vs2 based on density and Voigt stiffness.

        Parameters:
            density (float): The density of the material.
            voigt_stiffness (float): The Voigt stiffness of the material.
        """

        Plotter.plot_vp(voigt_stiffness, density)



import os
import json
import numpy as np
import pandas as pd
from tabulate import tabulate
from ..isotropy import Isotropy

class Material:
    def __init__(self, database_path=os.path.join(os.path.dirname(__file__), 'data/materials_database.json'), database_path2=os.path.join(os.path.dirname(__file__), 'data/derivatives_P.json'), database_path3=os.path.join(os.path.dirname(__file__),'data/derivatives_T.json')):
        """
        Initialize a Material object with data loaded from JSON files.

        Parameters:
        - database_path (str): The file path to the main materials database JSON file. Default is 'data/materials_database.json'.
        - database_path2 (str): The file path to the pressure derivatives JSON file. Default is 'data/derivatives_P.json'.
        - database_path3 (str): The file path to the temperature derivatives JSON file. Default is 'data/derivatives_T.json'.

        Returns:
        - None
        """
        self.materials_data = self.load_materials_data(database_path)
        self.pressure_deriv = self.load_pressure_deriv(database_path2)
        self.temperature_deriv = self.load_temperature_deriv(database_path3)
        # self.database_path_mat = os.path.join(os.path.dirname(__file__), 'data/materials_database.json')


    def load_materials_data(self, database_path):

        """
        Load materials data from a JSON file.

        Parameters:
        - database_path (str): The file path to the materials database JSON file.

        Returns:
        - dict: A dictionary containing the loaded materials data.
        """
        try:
            with open(database_path, 'r') as file:
                materials_data = json.load(file)
            return materials_data
        except FileNotFoundError:
            print("Database file not found.")
            return None
        except json.JSONDecodeError:
            print("Error decoding JSON data.")
            return None


    def load_pressure_deriv(self, database_path2):
        """
        Load pressure derivatives from a JSON file.

        Parameters:
        - database_path (str): The file path to the pressure derivatives JSON file.

        Returns:
        - dict: A dictionary containing the loaded pressure derivatives data.
        """
        try:
            with open(database_path2, 'r') as file:
                pressure_data = json.load(file)
            return pressure_data
        except FileNotFoundError:
            print("Database file not found.")
            return None
        except json.JSONDecodeError:
            print("Error decoding JSON data.")
            return None


    def load_temperature_deriv(self, database_path3):

        """
        Load temperature derivatives from a JSON file.

        Parameters:
        - database_path (str): The file path to the temperature derivatives JSON file.

        Returns:
        - dict: A dictionary containing the loaded temperature derivatives data.
        """
        try:
            with open(database_path3, 'r') as file:
                temperature_data = json.load(file)
            return temperature_data
        except FileNotFoundError:
            print("Database file not found.")
            return None
        except json.JSONDecodeError:
            print("Error decoding JSON data.")
            return None

    
    def load_density(self, phase, pressure = None, temperature = None):

        """
        Load the density of a material based on the given phase.
        
        Parameters:
        - phase (str): The phase of the material.
        - pressure (float or None): The pressure in MPa (optional).
        - temperature (float or None): The temperature in Kelvin (optional).
        
        Returns:
        - float or None: The density in g/cm³ or None if data not available.
        """
        if pressure is None or temperature is None:

            try:
                for material in self.materials_data:
                    try:
                        if material['Phase'] == phase:
                            return material["Density(g/cm3)"] * 1000
                    except KeyError:
                        print("Error: Density data not found or incorrect format.")
                        return None
                print(f"Error: Phase '{phase}' not found in materials data.")
                return None
            except AttributeError:
                print("Error: Materials data not available.")
                return None
            
        else:
            isotropy = Isotropy()
            density, aks, amu = isotropy.calculate_seismic_properties(phase, temperature=temperature, pressure=pressure, return_vp_vs_vbulk=False, return_aktout=False)
            return density

    

    def get_properties_by_phase(self, phase):

        """
        Get the properties of a material based on the given phase.
        
        Parameters:
        - phase (str): The phase of the material.
        
        Returns:
        - dict or None: The properties of the material or None if not found.
        """
        for material in self.materials_data:
            if material['Phase'] == phase:
                return material
        return None  # Return None if the phase is not found
    
    def get_pressure_properties_by_phase(self, phase):
        """
        Get the pressure properties of a material based on the given phase.
        
        Parameters:
        - phase (str): The phase of the material.
        
        Returns:
        - dict or None: The pressure properties or None if not found.
        """
        for material in self.pressure_deriv:
            if material['Phase'] == phase:
                return material
        return None  # Return None if the phase is not found
    
    def get_temperature_properties_by_phase(self, phase):
        """
        Get the temperature properties of a material based on the given phase.
        
        Parameters:
        - phase (str): The phase of the material.
        
        Returns:
        - dict or None: The temperature properties or None if not found.
        """

        for material in self.temperature_deriv:
            if material['Phase'] == phase:
                return material
        return None  # Return None if the phase is not found

    def get_voigt_matrix(self, phase):
        """
        Calculate and return the Voigt matrix of a material based on the given phase.
        
        Parameters:
        - phase (str): The phase of the material.
        
        Returns:
        - numpy array or None: The Voigt matrix or None if data not available.
        """
        material_properties = self.get_properties_by_phase(phase)

        if not material_properties:
            return None
        
        c11 = material_properties.get('C11', 0)
        c22 = material_properties.get('C22', 0)
        c33 = material_properties.get('C33', 0)
        c44 = material_properties.get('C44', 0)
        c55 = material_properties.get('C55', 0)
        c66 = material_properties.get('C66', 0)
        c12 = material_properties.get('C12', 0)
        c13 = material_properties.get('C13', 0)
        c23 = material_properties.get('C23', 0)
        c14 = material_properties.get('C14', 0)
        c15 = material_properties.get('C15', 0)
        c16 = material_properties.get('C16', 0)
        c24 = material_properties.get('C24', 0)
        c25 = material_properties.get('C25', 0)
        c26 = material_properties.get('C26', 0)
        c34 = material_properties.get('C34', 0)
        c35 = material_properties.get('C35', 0)
        c36 = material_properties.get('C36', 0)
        c45 = material_properties.get('C45', 0)
        c46 = material_properties.get('C46', 0)
        c56 = material_properties.get('C56', 0)

        voigt_matrix = np.array([
            [c11, c12, c13, c14, c15, c16],
            [c12, c22, c23, c24, c25, c26],
            [c13, c23, c33, c34, c35, c36],
            [c14, c24, c34, c44, c45, c46],
            [c15, c25, c35, c45, c55, c56],
            [c16, c26, c36, c46, c56, c66]
        ])

        # Make the matrix symmetric by copying the upper triangular part to the lower triangular part
        voigt_matrix = (voigt_matrix + voigt_matrix.T)/2

        return voigt_matrix
    
    def get_voigt_matrix_pressure(self, phase):
        material_properties = self.get_pressure_properties_by_phase(phase)

        if not material_properties:
            return None

        c11 = material_properties.get('C11', 0)
        c22 = material_properties.get('C22', 0)
        c33 = material_properties.get('C33', 0)
        c44 = material_properties.get('C44', 0)
        c55 = material_properties.get('C55', 0)
        c66 = material_properties.get('C66', 0)
        c12 = material_properties.get('C12', 0)
        c13 = material_properties.get('C13', 0)
        c23 = material_properties.get('C23', 0)
        c14 = material_properties.get('C14', 0)
        c15 = material_properties.get('C15', 0)
        c16 = material_properties.get('C16', 0)
        c24 = material_properties.get('C24', 0)
        c25 = material_properties.get('C25', 0)
        c26 = material_properties.get('C26', 0)
        c34 = material_properties.get('C34', 0)
        c35 = material_properties.get('C35', 0)
        c36 = material_properties.get('C36', 0)
        c45 = material_properties.get('C45', 0)
        c46 = material_properties.get('C46', 0)
        c56 = material_properties.get('C56', 0)

        voigt_matrix = np.array([
            [c11, c12, c13, c14, c15, c16],
            [c12, c22, c23, c24, c25, c26],
            [c13, c23, c33, c34, c35, c36],
            [c14, c24, c34, c44, c45, c46],
            [c15, c25, c35, c45, c55, c56],
            [c16, c26, c36, c46, c56, c66]
        ])

        # Make the matrix symmetric by copying the upper triangular part to the lower triangular part
        voigt_matrix = (voigt_matrix + voigt_matrix.T)/2

        return voigt_matrix
    
    def get_voigt_matrix_temperature(self, phase):
        material_properties = self.get_temperature_properties_by_phase(phase)

        if not material_properties:
            return None

        c11 = material_properties.get('C11', 0)
        c22 = material_properties.get('C22', 0)
        c33 = material_properties.get('C33', 0)
        c44 = material_properties.get('C44', 0)
        c55 = material_properties.get('C55', 0)
        c66 = material_properties.get('C66', 0)
        c12 = material_properties.get('C12', 0)
        c13 = material_properties.get('C13', 0)
        c23 = material_properties.get('C23', 0)
        c14 = material_properties.get('C14', 0)
        c15 = material_properties.get('C15', 0)
        c16 = material_properties.get('C16', 0)
        c24 = material_properties.get('C24', 0)
        c25 = material_properties.get('C25', 0)
        c26 = material_properties.get('C26', 0)
        c34 = material_properties.get('C34', 0)
        c35 = material_properties.get('C35', 0)
        c36 = material_properties.get('C36', 0)
        c45 = material_properties.get('C45', 0)
        c46 = material_properties.get('C46', 0)
        c56 = material_properties.get('C56', 0)

        voigt_matrix = np.array([
            [c11, c12, c13, c14, c15, c16],
            [c12, c22, c23, c24, c25, c26],
            [c13, c23, c33, c34, c35, c36],
            [c14, c24, c34, c44, c45, c46],
            [c15, c25, c35, c45, c55, c56],
            [c16, c26, c36, c46, c56, c66]
        ])

        # Make the matrix symmetric by copying the upper triangular part to the lower triangular part
        voigt_matrix = (voigt_matrix + voigt_matrix.T)/2

        return voigt_matrix
    

    
    def availablePhases(self):
        """
        Generate a formatted table listing available phases along with their crystal systems and primary phases.
        
        Returns:
        - str: A formatted table listing available phases.
        """
        headers = ["Phase", "Crystal System", "Primary Phase"]
        data = [(material['Phase'], material['Crystal System'], material['Primary Phase']) for material in self.materials_data]
        return tabulate(data, headers=headers, tablefmt="pretty")

    def availablePrimaryPhases(self):
        """
        Generate a formatted table listing available primary phases along with their crystal systems and phases.
        
        Returns:
        - str: A formatted table listing available primary phases.
        """
        headers = ["Primary Phase", "Crystal Systems", "Phases"]
        info = {}
        for material in self.materials_data:
            primary_phase = material['Primary Phase']
            crystal_system = material['Crystal System']
            phase = material['Phase']
            if primary_phase not in info:
                info[primary_phase] = {'Crystal Systems': [], 'Phases': []}
            info[primary_phase]['Crystal Systems'].append(crystal_system)
            info[primary_phase]['Phases'].append(phase)
        data = [(key, ", ".join(info[key]['Crystal Systems']), ", ".join(info[key]['Phases'])) for key in info]
        return tabulate(data, headers=headers, tablefmt="pretty")

    def availableCrystalSystems(self):
        """
        Generate a formatted table listing available crystal systems along with their associated phases.
        
        Returns:
        - str: A formatted table listing available crystal systems.
        """
        headers = ["Crystal System", "Phases"]
        crystal_systems = set(material['Crystal System'] for material in self.materials_data)
        info = {system: [] for system in crystal_systems}
        for material in self.materials_data:
            system = material['Crystal System']
            phase = material['Phase']
            info[system].append(phase)
        data = [(key, ", ".join(info[key])) for key in info]
        return tabulate(data, headers=headers, tablefmt="pretty")
    

    def voigthighPT(self, phase, PRESSURE=0, TEMP=300):
        """
        Perform calculations to obtain the Voigt matrix at high pressure and temperature conditions.
        
        Parameters:
        - phase (str): The phase of the material.
        - PRESSURE (float): The pressure in MPa (default 0).
        - TEMP (float): The temperature in Kelvin (default 300).
        
        Returns:
        - numpy array or None: The calculated Voigt matrix or None if data not available.
        """
        dCdT = self.get_voigt_matrix_temperature(phase)
        dCdP = self.get_voigt_matrix_pressure(phase)
        
        tensor_calc = self.get_voigt_matrix(phase) + (dCdT * (TEMP - 300) * 0.001) + (dCdP * PRESSURE)
        return tensor_calc
    
    def modalRock(self, rock, fraction, pressure, temperature, melt=0):
        """
        Calculate the average Voigt matrix and density of a rock composed of multiple minerals.
        
        Parameters:
        - rock (list): An array of minerals.
        - fraction (list): Fractions corresponding to each mineral in the rock.
        - pressure (float): The pressure in MPa.
        - temperature (float): The temperature in Kelvin.
        - melt (float): Amount of melting (default 0).
        
        Returns:
        - tuple: Average Voigt matrix as a numpy array and average density as a float.

        # Rock is an array of minerals and fraction is another array corresponding to the particular phase. eg. rock = ["Forsterite", "Diopside", "Enstatite"], fraction = [0.2, 0.5, 0.3]

        """
        cij = []
        rho = []
        for item in rock:
            cij.append(self.voigthighPT(item, pressure, temperature))
            rho.append(self.load_density(item, pressure, temperature))
        rho_average = np.sum(np.multiply(fraction, rho))
        cij_average = np.zeros((6, 6))

        for i in range(len(fraction)):
            cij_average += fraction[i] * cij[i]
        return cij_average, rho_average
    

    def modalRockFromExcel(self, excel_path, pressure, temperature, melt=0):
        """
        Calculate the average Voigt matrix and density of rocks from an Excel worksheet.
        
        Parameters:
        - excel_path (str): Path to the Excel worksheet.
        - pressure (float): The pressure in MPa.
        - temperature (float): The temperature in Kelvin.
        - melt (float): Amount of melting (default 0).
        
        Returns:
        - tuple: Average Voigt matrix as a numpy array and average density as a float.
        """

        df = pd.read_excel(excel_path)
        rock_data = df.values 

        cij_average = np.zeros((6, 6))
        rho_total = 0

        for rock, fraction in rock_data:
            cij = self.voigthighPT(rock, pressure, temperature)
            rho = self.load_density(rock, pressure, temperature)
            rho_total += fraction * rho
            cij_average += fraction * cij

        return cij_average, rho_total

            


#Adapted from Hacker et al 2003

import numpy as np
from scipy.optimize import root_scalar
import json
import os

class Isotropy:
    def __init__(self):
        """
        Initializes the Isotropy class by loading material data from a JSON file.

        Raises:
        FileNotFoundError: If the JSON file 'materials.json' is not found.
        json.JSONDecodeError: If there is an error decoding the JSON file.
        """
        try:
            json_file = 'materials.json'
            json_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data', json_file)
            with open(json_path, 'r') as f:
                self.data = json.load(f)
        except (FileNotFoundError, json.JSONDecodeError) as e:
            raise FileNotFoundError(f"Error loading JSON file '{json_file}': {e}")

    def get_available_phases(self):
        """
        Prints and returns a list of available phases from the loaded material data.

        Returns:
        List: A list of available phases.
        """
        try:
            for key in range(0,len(self.data)):
                print('######################Available Phases######################')
                print('Material id: ' + str(self.data[str(key)]['phase'] + '       ' + 'Material name: ' + str(self.data[str(key)]['name'])))
            return [self.data[key]['phase'] for key in self.data]
        except (KeyError, TypeError) as e:
            print(f"Error accessing data: {e}")
            return []

    def get_phase_constants(self, phase):
        """
        Retrieves phase constants for a given phase name or ID from the loaded material data.

        Parameters:
        phase (str): The name or ID of the phase.

        Returns:
        dict: Phase constants including ID, name, density, expansivity, bulk modulus, shear modulus, and more.
        """
        try:
            phase_data = next(item for item in self.data.values() if item.get('phase') == phase or item.get('name') == phase)
            if phase_data:
                return {
                    'id': phase_data['phase'],
                    'name': phase_data['name'],
                    'rho0': phase_data['density_298K(kg/m3)'],
                    'ao': phase_data['expansivity(/K)_a0'],
                    'akt0': phase_data['isothermal_bulk_modulus(P)_KT'],
                    'dkdp': phase_data['dKt/dP(KT_prime)'],
                    'amu0': phase_data['shear_modulus(Pa)_G'],
                    'dmudp': phase_data['dG/dP(G_prime)'],
                    'gam': phase_data['dlnG/dlnr_gamma'],
                    'grun': phase_data['first_gruinesen_parameter_gth'],
                    'delt': phase_data['second_gruinesen_parameter_dT']
                }
            else:
                raise KeyError(f"Phase '{phase}' not found in the data.")
        except (StopIteration, KeyError) as e:
            print(f"Error accessing phase constants: {e}")
            return None

    def calculate_seismic_properties(self, phase, temperature, pressure, ref_density = None, return_vp_vs_vbulk=False, return_aktout=False):
        """
        Calculate velocities at P,T for a mineral given parameters.
        
        Inputs:
        temperature (in K)
        pressure (in GPa)
        
        Outputs:
        
        density in kg/m3
        vp, vs, vbulk in km/s
        
        """
        
        if isinstance(temperature,(int,float)):
            if isinstance(pressure,(int,float)):
                method_calc = 'scalar'
            else:
                raise ValueError('While temperature is a scalar, pressure is an array.')
        else:
            if len(temperature) != len(pressure):
                raise ValueError('Length of the temperature and pressure arrays do not match!')
            else:
                method_calc = 'array'
        
        #Unit conversion from GPa
        
        pressure = pressure * 1e9 #conversion to GPa to Pa
        
        phase_constants = self.get_phase_constants(phase)

        if phase_constants:
            
            # Constants
            reference_temperature = 25.0
            absolute_temperature = 273.0
            tolerance = 1.e-8

            # input
            if ref_density == None:
                rho0 = phase_constants['rho0']
            else:
                rho0 = ref_density
            
            ao = phase_constants['ao']
            akt0 = phase_constants['akt0']
            dkdp = phase_constants['dkdp']
            amu0 = phase_constants['amu0']
            dmudp = phase_constants['dmudp']
            gam = phase_constants['gam']
            grun = phase_constants['grun']
            delt = phase_constants['delt']

            # first, integrate with temperature at P=0
            alpht = ao * (1. - 10. / np.sqrt(absolute_temperature + temperature))
            phi = ao * (temperature - reference_temperature) - 20. * ao * (np.sqrt(temperature + absolute_temperature) - np.sqrt(reference_temperature + absolute_temperature))
            density_t = rho0 * np.exp(-phi)

            akt_t = akt0 * np.exp(-delt * phi)  # KtT in Excel macro
            amu_t = amu0 * np.exp(-gam * phi)   # GT in Excel macro

            # second, integrate with pressure at T fixed
            f_guess = pressure / (3. * akt_t)
            f_guess2 = 0.
            akk = akt_t
            akk_prime = dkdp

            if method_calc == 'array':

                ff = []
                
                for i in range(0,len(temperature)):

                    res = root_scalar(self.pressure_function, args=(pressure[i], akk[i], akk_prime), bracket=[f_guess[i], f_guess2], xtol=tolerance)
                    if not res.converged:
                        raise ValueError("Root finding did not converge at array index: " + str(i))
                    ff_ = res.root
                    ff.append(ff_)

                ff = np.array(ff)

            else:

                res = root_scalar(self.pressure_function, args=(pressure, akk, akk_prime), bracket=[f_guess, f_guess2], xtol=tolerance)
                if not res.converged:
                    raise ValueError("Root finding did not converge.")
                ff = res.root
            
            dvol = (1. + 2. * ff) ** 1.5
            density = density_t * dvol
            ffac = (1. + 2. * ff) ** 2.5

            akt = akt_t * ffac * (1. - (5. - 3. * dkdp) * ff)
            amu = amu_t * ffac * (1. - ff * (5. - 3. * dmudp * akt_t / amu_t))

            # convert to adiabatic bulk modulus
            alphtp = alpht * dvol ** (-delt)
            aks = akt * (1. + alphtp * grun * (temperature + absolute_temperature))
            
            if return_vp_vs_vbulk == True:
                vbulk, vp, vs = self.calculate_velocities(density, aks, amu)
                if return_aktout == True:
                    return density, aks, amu, vp, vs, vbulk, akt
                else:
                    return density, aks, amu, vp, vs, vbulk
            else:
                if return_aktout == True:
                    return density, aks, amu, akt
                else:
                    return density, aks, amu
        else:
            raise ValueError(f"No phase constants available for '{phase}'.")


    def calculate_velocities(self, density, aks, amu):
        """
        Calculates velocities (vp, vs, vbulk) based on density, adiabatic bulk modulus (aks), and shear modulus (amu).

        Parameters:
        density (float or array): Density in kg/m^3.
        aks (float or array): Adiabatic bulk modulus in Pa.
        amu (float or array): Shear modulus in Pa.

        Returns:
        Tuple: Velocities vp, vs, vbulk in km/s.
        """
        try:
            vbulk = 0.001 * np.sqrt(aks / density)
            vp = 0.001 * np.sqrt((aks + 1.3333 * amu) / density)
            vs = 0.001 * np.sqrt(amu / density)
            return vbulk, vp, vs
        except (ZeroDivisionError, ValueError) as e:
            print(f"Error calculating velocities: {e}")

    def pressure_function(self, f, press, akk, akk_prime):
        """
        Helper function for pressure calculation in calculate_seismic_properties method.

        Parameters:
        f (float): Pressure factor.
        press (float): Pressure in Pa.
        akk (float): Adiabatic bulk modulus in Pa.
        akk_prime (float): Derivative of bulk modulus with respect to pressure.

        Returns:
        float: Pressure difference.
        """
        try:
            zeta = 0.75 * (4. - akk_prime)
            f1 = (1. + 2. * f) ** 2.5
            f2 = (1. - 2. * zeta * f)
            pdif = 3. * akk * f * f1 * f2 - press
            return pdif
        except ZeroDivisionError as e:
            print(f"Error in pressure function: {e}")
            
    def HashinShtrikmanBounds(self, phase_constant_list, fraction_list, temperature, pressure, fractions_matches_T = False):
    
        """
        Calculates Hashin-Shtrikman Bounds for the given parameters:
        
        Parameters:
        - phase_constant_list: list of phase constant instances gotten from set_modal_composition function.
        - fraction_list: list of fractions of each phase.
        - temperature: temperature array or scalar in Kelvin
        - pressure: pressure array or scalar in GPa
        
        Returns:
        - 3 lists of medium value, upper bound and lower bound, indexed respectively: 
            [Vbulk_medium,Vp_medium,Vs_medium], [Vbulk_upper,Vp_upper,Vs_upper], [Vbulk_lower, Vp_lower, Vs_lower]
        
        """
        aks_list = []
        amu_list = []
        
        if isinstance(temperature,(int,float)):
            if isinstance(pressure,(int,float)):
                method_calc = 'scalar'
            else:
                raise ValueError('While temperature is a scalar, pressure is an array.')
        else:
            if len(temperature) != len(pressure):
                raise ValueError('Length of the temperature and pressure arrays do not match!')
            else:
                method_calc = 'array'
                
                if len(fraction_list) == len(temperature):
                    fractions_matches_T = True
                else:
                    fractions_matches_T = False
                
        if method_calc == 'array':
            density_mix = np.zeros(len(temperature))
        else:
            density_mix = 0.0
        
        for phase_idx in range(0,len(phase_constant_list)):
        
            #calculating each phase moduli and density at T and P
            density, aks, amu = self.calculate_seismic_properties(phase_constant_list[phase_idx]['id'], temperature=temperature, pressure=pressure,
            return_vp_vs_vbulk=False, return_aktout=False)
            
            #calculating density of the mixture
            if fractions_matches_T == False:
                density_mix = density_mix + (density * fraction_list[phase_idx])
            else:
                density_mix = density_mix + (density * fraction_list[:,phase_idx])
            
            #appending moduli into lists
            aks_list.append(aks)
            amu_list.append(amu)
            
        
        aks_list = np.array(aks_list)
        amu_list = np.array(amu_list)
        
        #Minimum maximum bounds calculation for moduli
        aks_min_array = np.amin(aks_list, axis = 0)
        aks_max_array = np.amax(aks_list, axis = 0)
        
        amu_min_array = np.amin(amu_list, axis = 0)
        amu_max_array = np.amax(amu_list, axis = 0)
        
        zet_min = (amu_min_array / 6.0) * ((9.0*aks_min_array) + (8.0*amu_min_array)) / (aks_min_array + (2*amu_min_array))
        zet_max = (amu_max_array / 6.0) * ((9.0*aks_max_array) + (8.0*amu_max_array)) / (aks_max_array + (2*amu_max_array))
        
        if method_calc == 'array':
            skm = np.zeros(len(temperature))
            skp = np.zeros(len(temperature))
            smm = np.zeros(len(temperature))
            smp = np.zeros(len(temperature))
        else:
            skm = 0.0
            skp = 0.0
            smm = 0.0
            smp = 0.0
        
        for phase_idx in range(0,len(phase_constant_list)):
            if fractions_matches_T == False:
                frac_list_holder = fraction_list[phase_idx]
            else:
                frac_list_holder = fraction_list[:,phase_idx]
                
            skm = skm + (frac_list_holder / (aks_list[phase_idx] + (4*amu_min_array/3)))
            skp = skp + (frac_list_holder / (aks_list[phase_idx] + (4*amu_max_array/3)))
            smm = smm + (frac_list_holder / (amu_list[phase_idx] + zet_min))
            smp = smp + (frac_list_holder / (amu_list[phase_idx] + zet_max))
            
        khsm = (1.0 / skm) - (4.0*(amu_min_array/3.0))
        khsp = (1.0 / skp) - (4.0*(amu_max_array/3.0))
        mhsm = (1.0/smm) - zet_min
        mhsp = (1.0/smp) - zet_max
        
        rkkh = 0.5*(khsm+khsp)
        rkgh = 0.5*(mhsm+mhsp)
        
        Vbulk_medium, Vp_medium, Vs_medium = self.calculate_velocities(density = density_mix, aks = rkkh ,amu = rkgh)
        Vbulk_upper, Vp_upper, Vs_upper = self.calculate_velocities(density = density_mix, aks = khsp ,amu = mhsp)
        Vbulk_lower, Vp_lower, Vs_lower = self.calculate_velocities(density = density_mix, aks = khsm ,amu = mhsm)
        
        return [Vbulk_medium,Vp_medium,Vs_medium], [Vbulk_upper,Vp_upper,Vs_upper], [Vbulk_lower, Vp_lower, Vs_lower]
        
    def set_modal_composition(self, phase_list, fraction_list):
        
        """
        Prepares the input for HashinShtrikmanBounds method for the entered phase_list
        and fraction_list.
        
        Parameters:        
        - phase_list - list of phase ids to be included in the rock. 
            Example: ['fo', 'en'] -- Forsterite and Enstatite
        
        - fraction_list - list of fraction lists to be included in the rock
            Example: [0.8, 0.2] -- 0.8 fo and 0.2 en
            
        Returns:
        
        - phase_constant_list: 
            list containing the phase constats that can be inputted into HashinShtrikmanBounds method.
        - fraction_list: 
            np.ndarray list of fractions of entered minerals.
        """
        
        if type(fraction_list) != np.ndarray:
            fraction_list = np.array(fraction_list)
            
        all_sums = np.allclose(fraction_list.sum(axis = 1), 1.0, atol = 1e-6)
        
        if all_sums == False:
            raise ValueError('The values entered in fraction_list do not sum up to 1.')
        
        phase_constant_list = []
        
        for item in phase_list:
        
            phase_constant = self.get_phase_constants(phase = item)
            phase_constant_list.append(phase_constant)
                        
        return phase_constant_list, fraction_list
        
        


from .material import Material
from .ebsd import EBSD
from .tensor import Tensor
from .anisotropy import Anisotropy
from .anisotropy.plot_seismic_velocity import plot_vp_2d
from .isotropy import Isotropy


import numpy as np
from scipy.spatial.transform import Rotation
from joblib import Parallel, delayed
from tqdm import tqdm

def bunge_euler_rotation(phi1, Phi, phi2, angles):
    """
    Calculates the resultant Bunge Euler angles after applying a custom rotation.

    Args:
        phi1 (float): Initial Bunge Euler angle phi1 in degrees.
        Phi (float): Initial Bunge Euler angle Phi in degrees.
        phi2 (float): Initial Bunge Euler angle phi2 in degrees.
        angles (tuple): Custom rotation angles in degrees (alpha, beta, gamma).

    Returns:
        tuple: Resultant Bunge Euler angles (phi1, Phi, phi2) in degrees in the range [0, 360).
    """
    # Convert initial Bunge Euler angles to radians
    phi1_rad = np.deg2rad(phi1)
    Phi_rad = np.deg2rad(Phi)
    phi2_rad = np.deg2rad(phi2)

    # Create the initial rotation matrix from Bunge Euler angles
    initial_rot = Rotation.from_euler('zxz', [phi1_rad, Phi_rad, phi2_rad], degrees=False)

    # Create the custom rotation matrix
    alpha_rad, beta_rad, gamma_rad = [np.deg2rad(angle) for angle in angles]
    custom_rot = Rotation.from_euler('zxz', [alpha_rad, beta_rad, gamma_rad], degrees=False)

    # Calculate the resultant rotation matrix
    resultant_rot = initial_rot * custom_rot

    # Convert the resultant rotation matrix to Bunge Euler angles
    phi1_res, Phi_res, phi2_res = resultant_rot.as_euler('zxz', degrees=True)

    # Keep the angles in the range [0, 360)
    phi1_res = (phi1_res + 360) % 360
    Phi_res = (Phi_res + 360) % 360
    phi2_res = (phi2_res + 360) % 360

    return phi1_res, Phi_res, phi2_res

def apply_custom_rotation_to_dataframe(euler_df, angles):
    """
    Applies a custom rotation to each row of a DataFrame containing Bunge Euler angles.

    Args:
        euler_df (pd.DataFrame): DataFrame containing Bunge Euler angles.
        angles (tuple): Custom rotation angles in degrees (alpha, beta, gamma).

    Returns:
        pd.DataFrame: DataFrame with updated Bunge Euler angles after rotation.
    """
    def bunge_euler_rotation_parallel(row):
        """
        Calculates the resultant Bunge Euler angles after applying a custom rotation in parallel.

        Args:
            row (pd.Series): Row of DataFrame containing Euler angles.

        Returns:
            tuple: Resultant Bunge Euler angles (phi1, Phi, phi2) in degrees in the range [0, 360).
        """
        # Extract Euler angles from the row
        phi1, Phi, phi2 = row['Euler1'], row['Euler2'], row['Euler3']
        
        # Call the original rotation function
        return bunge_euler_rotation(phi1, Phi, phi2, angles)

    # Calculate the number of iterations
    num_iterations = len(euler_df)

    # Parallelize the process and show progress in percent
    results = Parallel(n_jobs=-1)(
        delayed(bunge_euler_rotation_parallel)(row)
        for _, row in tqdm(euler_df.iterrows(), total=num_iterations)
    )

    # Update the DataFrame with the new Euler angles
    euler_df[['Euler1', 'Euler2', 'Euler3']] = results

    return euler_df

def apply_custom_rotation_to_dataframe_noxy(df, angles):
    """
    Applies a custom rotation to each row of a DataFrame containing Bunge Euler angles, X, and Y coordinates.

    Args:
        df (pd.DataFrame): DataFrame containing Bunge Euler angles, X, and Y coordinates.
        angles (tuple): Custom rotation angles in degrees (alpha, beta, gamma).

    Returns:
        pd.DataFrame: DataFrame with updated Bunge Euler angles, X, and Y coordinates after rotation.
    """
    def rotation_parallel(row):
        """
        Calculates the resultant Bunge Euler angles, X, and Y coordinates after applying a custom rotation in parallel.

        Args:
            row (pd.Series): Row of DataFrame containing Euler angles, X, and Y coordinates.

        Returns:
            tuple: Resultant Bunge Euler angles (phi1, Phi, phi2) in degrees in the range [0, 360),
                   rotated X and Y coordinates.
        """
        # Extract Euler angles, X, and Y coordinates from the row
        phi1, Phi, phi2 = row['Euler1'], row['Euler2'], row['Euler3']
        x, y = row['X'], row['Y']
        
        # Apply rotation to Euler angles
        phi1_res, Phi_res, phi2_res = bunge_euler_rotation(phi1, Phi, phi2, angles)
        
        # Apply rotation to X and Y coordinates
        rotated_coords = Rotation.from_euler('zxz', [np.deg2rad(angle) for angle in angles], degrees=False).apply([x, y, 0])
        
        return phi1_res, Phi_res, phi2_res, rotated_coords[0], rotated_coords[1]

    # Calculate the number of iterations
    num_iterations = len(df)

    # Parallelize the process and show progress in percent
    results = Parallel(n_jobs=-1)(
        delayed(rotation_parallel)(row)
        for _, row in tqdm(df.iterrows(), total=num_iterations)
    )

    # Update the DataFrame with the new Euler angles, X, and Y coordinates
    df[['Euler1', 'Euler2', 'Euler3', 'X', 'Y']] = results

    return df



from ..material import Material
from ..anisotropy import Anisotropy

import numpy as np

def modalAnisotropy(material, fraction, pressure, temperature):
    voigtMatrixTotal = []
    density = []
    for i in range(len(material)):
        material_instance = Material()
        phase = 'Diopside'
        voigtMatrix = material_instance.voigthighPT(material[i], PRESSURE = pressure, TEMP = temperature)
        voigtMatrixTotal.append(voigtMatrix)
        density.append(material_instance.load_density(material[i]))

    aggregateStiffnessTensor = np.prod(np.array(voigtMatrixTotal), np.array(fraction))
    density_aggregate = np.prod(np.array(density), np.array(fraction))
    
    anisotropy_instance = Anisotropy(stiffness_matrix=aggregateStiffnessTensor, density=density_aggregate)






import pandas as pd
import pickle

def saveEBSDdf(dataframe, file_path):
    """Save a DataFrame to a file."""
    dataframe.to_csv(file_path, index=False)

def loadEBSDdf(file_path):
    """Load a DataFrame from a file."""
    return pd.read_csv(file_path)

def savePhaseName(lst, file_path):
    """Save a list to a file using pickle."""
    with open(file_path, 'wb') as f:
        pickle.dump(lst, f)

def loadPhaseName(file_path):
    """Load a list from a file using pickle."""
    with open(file_path, 'rb') as f:
        return pickle.load(f)



import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import cartopy
import numpy as np
from scipy.interpolate import griddata

def plot_data_with_interpolation_in_defined_box(lon_points, lat_points, values, interp_box, method='linear', full_map=False):
    """
    Plot interpolated values within a defined box on a map of South Australia, with an option for full map display.
    
    Parameters:
    - lon_points: Array-like, longitudes of the data points.
    - lat_points: Array-like, latitudes of the data points.
    - values: Array-like, values at the given data points.
    - interp_box: Tuple, defining the interpolation box (left_lon, bottom_lat, right_lon, top_lat).
    - method: String, interpolation method ('linear', 'nearest', etc.).
    - full_map: Boolean, if True, the full global map is displayed; otherwise, focus on the data points' region.
    """
    left_lon, bottom_lat, right_lon, top_lat = interp_box
    
    lon_grid, lat_grid = np.meshgrid(
        np.linspace(left_lon, right_lon, 100),
        np.linspace(bottom_lat, top_lat, 100)
    )
    
    interp_values = griddata(
        (lon_points, lat_points), values, (lon_grid, lat_grid), method=method
    )
    
    fig, ax = plt.subplots(figsize=(10, 10), subplot_kw={'projection': ccrs.PlateCarree()})
    
    if full_map:
        ax.set_global()
    else:
        ax.set_extent([left_lon, right_lon, bottom_lat, top_lat], crs=ccrs.PlateCarree())
    
    ax.coastlines(resolution='10m')
    ax.add_feature(cartopy.feature.LAND)
    ax.add_feature(cartopy.feature.OCEAN)
    
    contourf = ax.contourf(lon_grid, lat_grid, interp_values, transform=ccrs.PlateCarree(), cmap='viridis', alpha=0.6)
    plt.colorbar(contourf, ax=ax, shrink=0.5, label='Interpolated Value')
    
    ax.plot(lon_points, lat_points, 'ro', markersize=5, transform=ccrs.Geodetic(), label='Data Points')
    
    plt.legend()
    plt.title("Interpolated Data Map")
    plt.show()






import numpy as np
from santex import Material
import pandas as pd

def calc_melt_density(weight_per, pressure, temperature):
    """
    Calculates the density and component density of a melt based on its composition.

    Parameters:
        weight_per (list or dict): List or dictionary representing the weight percentage of various oxides in the melt. 
                                   If a dictionary is provided, it should have oxide names as keys and corresponding weight percentages as values.
        pressure (float): Pressure of the melt in bars.
        temperature (float): Temperature of the melt in Kelvin.

    Returns:
        density (float): Density of the melt.
        component_density (array): Component-wise density of the melt.

    Example Usage: 
    oxide = ['sio2', 'tio2', 'al2o3', 'fe2o3', 'feo', 'mgo', 'cao', 'na2o', 'k2o', 'h2o']
    weight_per = [40, 1.84, 13.7, 2.7, 9.57, 6.67, 11.5, 2.68, 0.25, 10] or weight_per can be a dictionary like {'sio2': 40, 'tio2': 1.84, 'feo': 9.57, 'mgo': 6.67, 'cao': 11.5, 'na2o': 2.68, 'k2o': 0.25, 'h2o': 10}
    """
    if type(weight_per) == list:
        oxide = ['sio2', 'tio2', 'al2o3', 'fe2o3', 'feo', 'mgo', 'cao', 'na2o', 'k2o', 'h2o']
        total = np.sum(weight_per)
        normalised_wt_per = np.array(weight_per)/total*100
        mol_weight = [60.0855, 79.88, 101.96, 159.69, 71.85, 40.3, 56.08, 61.98, 94.2, 18]
        moles = normalised_wt_per/mol_weight
        total_moles = np.sum(moles)
        mole_fraction = moles/total_moles
        molar_volume = np.array([26.86, 28.32, 37.42, 41.5, 12.68, 12.02, 16.9, 29.65, 47.28, 22.9])
        tref = np.array([1773, 1773, 1773, 1723, 1723, 1773, 1773, 1773, 1773, 1273])
        dVdT = np.array([0, 0.00724, 0.00262, 0, 0.00369, 0.00327, 0.00374, 0.00768, 0.01208, 0.0095])
        dVdP = np.array([-0.000189, -0.000231, -0.000226, -0.000253, -0.000045, 0.000027, 0.000034, -0.00024, -0.000675, -0.00032])
        temp_k = np.array([temperature]*len(oxide))
        press_b = np.array([pressure]*len(oxide))
        component_density = (mole_fraction*mol_weight)/(molar_volume+(dVdT*(temp_k-tref))+(dVdP*(press_b-1)))
        div_vliq = (molar_volume+(dVdT*(temp_k-tref))+(dVdP*(press_b-1)))*mole_fraction
        vliq = np.sum(div_vliq)
        div_x_mol_wt = mole_fraction*mol_weight
        sum_div_x_mol_wt = np.sum(div_x_mol_wt)
        density = sum_div_x_mol_wt/vliq
        return density, component_density
    else:
        oxide_dict = weight_per
        oxides = ['sio2', 'tio2', 'al2o3', 'fe2o3', 'feo', 'mgo', 'cao', 'na2o', 'k2o', 'h2o']
        mol_weight = np.array([60.0855, 79.88, 101.96, 159.69, 71.85, 40.3, 56.08, 61.98, 94.2, 18])
        
        weight_per = {oxide: 0 for oxide in oxides}
        weight_per.update(oxide_dict)
        
        weight_per_list = [weight_per[oxide] for oxide in oxides]
        
        total = sum(weight_per_list)
        normalised_wt_per = np.array(weight_per_list) / total * 100
        moles = normalised_wt_per / mol_weight
        total_moles = np.sum(moles)
        mole_fraction = moles / total_moles
        molar_volume = np.array([26.86, 28.32, 37.42, 41.5, 12.68, 12.02, 16.9, 29.65, 47.28, 22.9])
        tref = np.array([1773, 1773, 1773, 1723, 1723, 1773, 1773, 1773, 1773, 1273])
        dVdT = np.array([0, 0.00724, 0.00262, 0, 0.00369, 0.00327, 0.00374, 0.00768, 0.01208, 0.0095])
        dVdP = np.array([-0.000189, -0.000231, -0.000226, -0.000253, -0.000045, 0.000027, 0.000034, -0.00024, -0.000675, -0.00032])
        temp_k = np.array([temperature] * len(oxides))
        press_b = np.array([pressure] * len(oxides))
        component_density = (mole_fraction * mol_weight) / (molar_volume + (dVdT * (temp_k - tref)) + (dVdP * (press_b - 1)))
        div_vliq = (molar_volume + (dVdT * (temp_k - tref)) + (dVdP * (press_b - 1))) * mole_fraction
        vliq = np.sum(div_vliq)
        div_x_mol_wt = mole_fraction * mol_weight
        sum_div_x_mol_wt = np.sum(div_x_mol_wt)
        density = sum_div_x_mol_wt / vliq
        
        return density, component_density

def calc_melt_density_from_excel(input_file, output_file):
    """
    Calculates melt densities for samples listed in an input Excel file and saves the results to an output Excel file.

    Parameters:
        input_file (str): Path to the input Excel file containing sample data.
        output_file (str): Path to save the output Excel file with calculated melt densities.

    Sample excel file:
        Sample Name	Pressure (bars)	Temperature (K)	    SiO2	TiO2	Al2O3	Fe2O3	FeO	    MgO	    CaO	    Na2O	K2O	    H2O
        sample_wir_01	500	            1273	        45.5	2.1	    15.8	3.4	    8.2	    4.9	    10.2	3.3	    0.6	    5.5
        sample_wir_02	500	            1273	        49.7	1.9	    14.3	2.8	7   .1	    5.6	    9.8	    4.1	    0.7	    4.0

    """
    df = pd.read_excel(input_file)
    
    results_df = pd.DataFrame(columns=['Sample Name', 'Density', 'Component Density'])

    for index, row in df.iterrows():
        sample_name = row['Sample Name']
        pressure = row['Pressure (bars)']
        temperature = row['Temperature (K)']
        
        oxide_dict = row[3:].to_dict()
        
        density, component_density = calc_melt_density(oxide_dict, pressure, temperature)
        
        results_df = results_df.append({
            'Sample Name': sample_name, 
            'Density': density, 
            'Component Density': component_density
        }, ignore_index=True)

    results_df.to_excel(output_file, index=False)



def modalRock(rock, fraction, pressure, temperature, melt=0, weight_per = []):

    """
    Calculates the average elastic constants and density of a rock mixture.

    Parameters:
        rock (list): An array of mineral names constituting the rock mixture.
        fraction (list): An array corresponding to the fractional composition of each mineral in the rock.
        pressure (float): Pressure of the rock mixture in GPa.
        temperature (float): Temperature of the rock mixture in Kelvin.
        melt (float, optional): Fraction of melt in the rock mixture. Set to a non-zero value if there is a melt component. Default is 0.
        weight_per (list, optional): Weight percentage of oxides in the melt component, required if `melt` is non-zero. Default is [].

    Returns:
        cij_average (array): Average elastic constants of the rock mixture.
        rho_average (float): Average density of the rock mixture.

    Example Usage:
        Rock is an array of minerals and fraction is another array corresponding to the particular phase. 
        eg. rock = ["Forsterite", "Diopside", "Enstatite"], fraction = [0.2, 0.5, 0.3]
    """
    if melt:
        density, comp_density = calc_melt_density(weight_per, pressure*10000, temperature)
        voigt_melt = np.array([[16.1,   15.967,  15.967,   0.,      0,   0.   ],
                                [ 15.967, 16.1,   15.967,    0.,      0,   0.   ],
                                [ 15.967,  15.967,  16.1,    0.,     0,    0.   ],
                                [  0.,      0.,      0.,     0.01,    0.,      0],
                                [  0,   0,  0,    0.,     0.01,    0.   ],
                                [  0.,      0.,     0.,      0,   0.,     0.01 ]])
        cij = []
        rho = []
        material = Material()
        for item in rock:
            cij.append(material.voigthighPT(item, pressure, temperature))
            rho.append(material.load_density(item, pressure, temperature))
        rho_average = np.sum(np.multiply(fraction, rho))
        cij_average = np.zeros((6, 6))

        for i in range(len(fraction)):
            cij_average += fraction[i] * cij[i]

        cij_average_melt = ((cij_average*(1-melt))+(melt*(voigt_melt)))
        density_average_melt = rho_average*(1-melt) + density*melt
        return cij_average_melt, density_average_melt
    cij = []
    rho = []
    material = Material()
    for item in rock:
        cij.append(material.voigthighPT(item, pressure, temperature))
        rho.append(material.load_density(item, pressure, temperature))
    rho_average = np.sum(np.multiply(fraction, rho))
    cij_average = np.zeros((6, 6))

    for i in range(len(fraction)):
        cij_average += fraction[i] * cij[i]
    return cij_average, rho_average









import pandas as pd
import numpy as np
from joblib import Parallel, delayed

def euler_to_quaternion(phi, theta, psi):
    """
    Convert Euler angles (in radians) to a quaternion.

    Parameters:
        phi (float): Euler angle phi in radians.
        theta (float): Euler angle theta in radians.
        psi (float): Euler angle psi in radians.

    Returns:
        np.ndarray: A quaternion representing the input Euler angles.    
    """
    c1 = np.cos(phi / 2)
    s1 = np.sin(phi / 2)
    c2 = np.cos(theta / 2)
    s2 = np.sin(theta / 2)
    c3 = np.cos(psi / 2)
    s3 = np.sin(psi / 2)

    q0 = c1 * c2 * c3 + s1 * s2 * s3
    q1 = s1 * c2 * c3 - c1 * s2 * s3
    q2 = c1 * s2 * c3 + s1 * c2 * s3
    q3 = c1 * c2 * s3 - s1 * s2 * c3

    return np.array([q0, q1, q2, q3])

def quaternion_to_angle(q1, q2):
    """
    Calculate the misorientation angle between two quaternions.

    Parameters:
        q1 (np.ndarray): First quaternion.
        q2 (np.ndarray): Second quaternion.

    Returns:
        float: The misorientation angle between the two quaternions in degrees.
    """
    dot_product = np.dot(q1, q2)
    angle_radians = 2 * np.arccos(np.abs(dot_product))
    return np.degrees(angle_radians)

def misorientation_angle(euler1, euler2):
    """
    Calculate the misorientation angle between two sets of Euler angles.

    Parameters:
        euler1 (list or np.ndarray): First set of Euler angles in radians.
        euler2 (list or np.ndarray): Second set of Euler angles in radians.

    Returns:
        float: The misorientation angle between the two sets of Euler angles in degrees.    
    """
    q1 = euler_to_quaternion(*np.radians(euler1))
    q2 = euler_to_quaternion(*np.radians(euler2))
    return quaternion_to_angle(q1, q2)

def assign_to_grains_parallel(df, threshold):
    """
        Assign grains to rows in a DataFrame based on misorientation angle threshold.

    Parameters:
        df (pd.DataFrame): Input DataFrame containing Euler angles for each row.
        threshold (float): Misorientation angle threshold in degrees.

    Returns:
        dict: A dictionary mapping row indices to assigned grain indices.
    """
    grains = []
    grain_indices = {}

    total_rows = len(df)
    processed_rows = 0

    def process_row(row):
        nonlocal grains, grain_indices, processed_rows
        euler_angles = row[['Euler1', 'Euler2', 'Euler3']].values
        assigned = False
        for grain_idx, grain_euler in enumerate(grains):
            misorientation = misorientation_angle(euler_angles, grain_euler)
            if misorientation <= threshold:
                grain_indices[row.name] = grain_idx
                grains[grain_idx] = np.mean([grain_euler, euler_angles], axis=0)
                assigned = True
                break
        if not assigned:
            grain_indices[row.name] = len(grains)
            grains.append(euler_angles)

        processed_rows += 1
        progress = (processed_rows / total_rows) * 100
        print(f"Progress: {progress:.2f}%\r", end='')

    Parallel(n_jobs=-1, prefer="threads")(delayed(process_row)(row) for _, row in df.iterrows())

    return grain_indices




import numpy as np
from vtk import (
    vtkStructuredGrid, vtkPoints, vtkFloatArray,
    vtkStructuredGridGeometryFilter, vtkPolyDataMapper,
    vtkActor, vtkRenderer, vtkRenderWindow,
    vtkRenderWindowInteractor
)
from ..tensor.tensor import Tensor
from .utils import christoffel_tensor, wave_property

tensor = Tensor()
class Plotter:
    """
    A class for plotting seismic wave velocities in 3D using VTK.

    Attributes:
    - None
    """
    @classmethod
    def plot_vs_splitting(cls, c, rho):
        """
        Plot the difference between S-wave velocities (vs1 - vs2) in 3D using VTK.

        Parameters:
        - c (array): The fourth-rank stiffness tensor in Voigt notation.
        - rho (float): Density of the material.

        Returns:
        - None
        """
        cijkl = tensor.voigt_to_tensor(c)

        theta = np.linspace(0, np.pi, 100)
        phi = np.linspace(0, 2 * np.pi, 100)
        theta, phi = np.meshgrid(theta, phi)

        x = np.sin(theta) * np.cos(phi)
        y = np.sin(theta) * np.sin(phi)
        z = np.cos(theta)

        vp = np.zeros((100, 100))
        vs1 = np.zeros((100, 100))
        vs2 = np.zeros((100, 100))

        for i in range(len(theta)):
            for j in range(len(phi)):
                n = np.array([np.sin(theta[i][j]) * np.cos(phi[i][j]),
                              np.sin(theta[i][j]) * np.sin(phi[i][j]),
                              np.cos(theta[i][j])])
                tik = christoffel_tensor(cijkl, n)
                wave_moduli, _ = wave_property(tik)
                vp[i][j] = np.sqrt(wave_moduli[0] / rho)
                vs1[i][j] = np.sqrt(wave_moduli[1] / rho)
                vs2[i][j] = np.sqrt(wave_moduli[2] / rho)

        grid = vtkStructuredGrid()
        points = vtkPoints()
        values = vtkFloatArray()

        for i in range(len(theta)):
            for j in range(len(phi)):
                points.InsertNextPoint(x[i][j] * (vs1[i][j] - vs2[i][j]),
                                       y[i][j] * (vs1[i][j] - vs2[i][j]),
                                       z[i][j] * (vs1[i][j] - vs2[i][j]))
                values.InsertNextValue(vp[i][j])

        grid.SetDimensions(100, 100, 1)
        grid.SetPoints(points)
        grid.GetPointData().SetScalars(values)

        geometry_filter = vtkStructuredGridGeometryFilter()
        geometry_filter.SetInputData(grid)

        mapper = vtkPolyDataMapper()
        mapper.SetInputConnection(geometry_filter.GetOutputPort())
        mapper.SetScalarRange(np.min(vp), np.max(vp))

        actor = vtkActor()
        actor.SetMapper(mapper)

        renderer = vtkRenderer()
        renderer.AddActor(actor)
        renderer.SetBackground(1, 1, 1)

        render_window = vtkRenderWindow()
        render_window.AddRenderer(renderer)

        interactor = vtkRenderWindowInteractor()
        interactor.SetRenderWindow(render_window)

        interactor.Initialize()
        interactor.Start()

    @classmethod
    def plot_vp(cls, c, rho):
        """
        Plot the P-wave velocities (vp) in 3D using VTK.

        Parameters:
        - c (array): The fourth-rank stiffness tensor in Voigt notation.
        - rho (float): Density of the material.

        Returns:
        - None
        """
        cijkl = tensor.voigt_to_tensor(c)

        theta = np.linspace(0, np.pi, 100)
        phi = np.linspace(0, 2 * np.pi, 100)
        theta, phi = np.meshgrid(theta, phi)

        x = np.sin(theta) * np.cos(phi)
        y = np.sin(theta) * np.sin(phi)
        z = np.cos(theta)

        vp = np.zeros((100, 100))
        vs1 = np.zeros((100, 100))
        vs2 = np.zeros((100, 100))

        for i in range(len(theta)):
            for j in range(len(phi)):
                n = np.array([np.sin(theta[i][j]) * np.cos(phi[i][j]),
                              np.sin(theta[i][j]) * np.sin(phi[i][j]),
                              np.cos(theta[i][j])])
                tik = christoffel_tensor(cijkl, n)
                wave_moduli, _ = wave_property(tik)
                vp[i][j] = np.sqrt(wave_moduli[0] / rho)
                vs1[i][j] = np.sqrt(wave_moduli[1] / rho)
                vs2[i][j] = np.sqrt(wave_moduli[2] / rho)

        grid = vtkStructuredGrid()
        points = vtkPoints()
        values = vtkFloatArray()

        for i in range(len(theta)):
            for j in range(len(phi)):
                points.InsertNextPoint(x[i][j] * (vp[i][j]),
                                       y[i][j] * (vp[i][j]),
                                       z[i][j] * (vp[i][j]))
                values.InsertNextValue(vp[i][j])

        grid.SetDimensions(100, 100, 1)
        grid.SetPoints(points)
        grid.GetPointData().SetScalars(values)

        geometry_filter = vtkStructuredGridGeometryFilter()
        geometry_filter.SetInputData(grid)

        mapper = vtkPolyDataMapper()
        mapper.SetInputConnection(geometry_filter.GetOutputPort())
        mapper.SetScalarRange(np.min(vp), np.max(vp))

        actor = vtkActor()
        actor.SetMapper(mapper)

        renderer = vtkRenderer()
        renderer.AddActor(actor)
        renderer.SetBackground(1, 1, 1)

        render_window = vtkRenderWindow()
        render_window.AddRenderer(renderer)

        interactor = vtkRenderWindowInteractor()
        interactor.SetRenderWindow(render_window)

        interactor.Initialize()
        interactor.Start()

    @classmethod
    def plot_vs1(cls, c, rho):
        """
        Plot the S-wave velocities with higher velocity (vs1) in 3D using VTK.

        Parameters:
        - c (array): The fourth-rank stiffness tensor in Voigt notation.
        - rho (float): Density of the material.

        Returns:
        - None
        """
        cijkl = tensor.voigt_to_tensor(c)

        theta = np.linspace(0, np.pi, 100)
        phi = np.linspace(0, 2 * np.pi, 100)
        theta, phi = np.meshgrid(theta, phi)

        x = np.sin(theta) * np.cos(phi)
        y = np.sin(theta) * np.sin(phi)
        z = np.cos(theta)

        vp = np.zeros((100, 100))
        vs1 = np.zeros((100, 100))
        vs2 = np.zeros((100, 100))

        for i in range(len(theta)):
            for j in range(len(phi)):
                n = np.array([np.sin(theta[i][j]) * np.cos(phi[i][j]),
                              np.sin(theta[i][j]) * np.sin(phi[i][j]),
                              np.cos(theta[i][j])])
                tik = christoffel_tensor(cijkl, n)
                wave_moduli, _ = wave_property(tik)
                vp[i][j] = np.sqrt(wave_moduli[0] / rho)
                vs1[i][j] = np.sqrt(wave_moduli[1] / rho)
                vs2[i][j] = np.sqrt(wave_moduli[2] / rho)

        grid = vtkStructuredGrid()
        points = vtkPoints()
        values = vtkFloatArray()

        for i in range(len(theta)):
            for j in range(len(phi)):
                points.InsertNextPoint(x[i][j] * (vs1[i][j]),
                                       y[i][j] * (vs1[i][j]),
                                       z[i][j] * (vs1[i][j]))
                values.InsertNextValue(vp[i][j])

        grid.SetDimensions(100, 100, 1)
        grid.SetPoints(points)
        grid.GetPointData().SetScalars(values)

        geometry_filter = vtkStructuredGridGeometryFilter()
        geometry_filter.SetInputData(grid)

        mapper = vtkPolyDataMapper()
        mapper.SetInputConnection(geometry_filter.GetOutputPort())
        mapper.SetScalarRange(np.min(vp), np.max(vp))

        actor = vtkActor()
        actor.SetMapper(mapper)

        renderer = vtkRenderer()
        renderer.AddActor(actor)
        renderer.SetBackground(1, 1, 1)

        render_window = vtkRenderWindow()
        render_window.AddRenderer(renderer)

        interactor = vtkRenderWindowInteractor()
        interactor.SetRenderWindow(render_window)

        interactor.Initialize()
        interactor.Start()
    @classmethod
    def plot_vp(cls, c, rho):

        """
        Plot the P-wave velocities (vp) in 3D using VTK.

        This method calculates and visualizes the P-wave velocities (vp) for a given fourth-rank stiffness tensor (c)
        and material density (rho) using the Visualization Toolkit (VTK).

        Parameters:
        - c (array): The fourth-rank stiffness tensor in Voigt notation.
        - rho (float): Density of the material.

        Returns:
        - None

        Usage Example:
        >>> c = np.array([[323.70, 66.40, 71.60, 0.000, 0.000, 0.000],
        ...               [66.40, 197.60, 75.60, 0.000, 0.000, 0.000],
        ...               [71.60, 75.60, 235.10, 0.000, 0.000, 0.000],
        ...               [0.000, 0.000, 0.000, 64.62, 0.000, 0.000],
        ...               [0.000, 0.000, 0.000, 0.000, 78.05, 0.000],
        ...               [0.000, 0.000, 0.000, 0.000, 0.000, 79.04]]) * 1e9
        >>> Plotter.plot_vp(c, 3310)
        """

        cijkl = tensor.voigt_to_tensor(c)

        theta = np.linspace(0, np.pi, 100)
        phi = np.linspace(0, 2 * np.pi, 100)
        theta, phi = np.meshgrid(theta, phi)

        x = np.sin(theta) * np.cos(phi)
        y = np.sin(theta) * np.sin(phi)
        z = np.cos(theta)

        vp = np.zeros((100, 100))
        vs1 = np.zeros((100, 100))
        vs2 = np.zeros((100, 100))

        for i in range(len(theta)):
            for j in range(len(phi)):
                n = np.array([np.sin(theta[i][j]) * np.cos(phi[i][j]),
                              np.sin(theta[i][j]) * np.sin(phi[i][j]),
                              np.cos(theta[i][j])])
                tik = christoffel_tensor(cijkl, n)
                wave_moduli, _ = wave_property(tik)
                vp[i][j] = np.sqrt(wave_moduli[0] / rho)
                vs1[i][j] = np.sqrt(wave_moduli[1] / rho)
                vs2[i][j] = np.sqrt(wave_moduli[2] / rho)

        grid = vtkStructuredGrid()
        points = vtkPoints()
        values = vtkFloatArray()

        for i in range(len(theta)):
            for j in range(len(phi)):
                points.InsertNextPoint(x[i][j] * (vs2[i][j]),
                                       y[i][j] * (vs2[i][j]),
                                       z[i][j] * (vs2[i][j]))
                values.InsertNextValue(vp[i][j])

        grid.SetDimensions(100, 100, 1)
        grid.SetPoints(points)
        grid.GetPointData().SetScalars(values)

        geometry_filter = vtkStructuredGridGeometryFilter()
        geometry_filter.SetInputData(grid)

        mapper = vtkPolyDataMapper()
        mapper.SetInputConnection(geometry_filter.GetOutputPort())
        mapper.SetScalarRange(np.min(vp), np.max(vp))

        actor = vtkActor()
        actor.SetMapper(mapper)

        renderer = vtkRenderer()
        renderer.AddActor(actor)
        renderer.SetBackground(1, 1, 1)

        render_window = vtkRenderWindow()
        render_window.AddRenderer(renderer)

        interactor = vtkRenderWindowInteractor()
        interactor.SetRenderWindow(render_window)

        interactor.Initialize()
        interactor.Start()



import pandas as pd
import csv

class Ctf:
    def __init__(self, filename) -> None:
        self._filename = filename

    def phases_info(self):

        lookup_string = ["Euler angles refer to Sample Coordinate system (CS0)!", "Phases"]

        with open(self._filename, 'r') as f:
            reader = csv.reader(f, delimiter='\t')

            for i, line in enumerate(reader):
                line_str = '\t'.join(line)

                if lookup_string[0] in line_str:

                    fields = line_str.split('\t')

                    self.mag = fields[fields.index('Mag') + 1]
                    self.coverage = fields[fields.index('Coverage') + 1]
                    self.device = fields[fields.index('Device') + 1]
                    self.kv = fields[fields.index('KV') + 1]
                    self.tilt_angle = fields[fields.index('TiltAngle') + 1]

                    line_number = i + 1


                if lookup_string[1] in line_str:
                    fields = line_str.split('\t')
                    self.nphases = int(fields[fields.index('Phases') + 1])
                    phases_start_line = i + 1


        phases_columns = ["cell_length_a", "cell_length_b", "cell_length_c", "cell_angle_alpha", "cell_angle_beta", "cell_angle_gamma", "phase", "symmetry", "nan1", "nan2", "nan3", "author"]
        self.phases = pd.read_csv(self._filename, delimiter='[;\t]', skiprows=phases_start_line, nrows=self.nphases, names=phases_columns, index_col=None, engine="python")
        self.phases.index = self.phases.index + 1

        return self.nphases, self.mag, self.coverage, self.device, self.kv, self.tilt_angle, self.phases
            

    def get_data(self):
        with open(self._filename, 'r') as file:
            lines = file.readlines()

        header_start = lines.index('JobMode\tGrid\n')
        data_start = lines.index('Phase\tX\tY\tBands\tError\tEuler1\tEuler2\tEuler3\tMAD\tBC\tBS\n')
        header_data = pd.read_csv(self._filename, delimiter='\t', skiprows=header_start+1, nrows=7, names=["info", "value"])
        data = pd.read_csv(self._filename, delimiter='\t', skiprows=data_start)
        return data, header_data

    def get_phases_numbers(self):
        return self.get_data



import numpy as np
import math
import matplotlib.pyplot as plt
from .utils import christoffel_tensor, wave_property

def phase_velocity(cijkl, rho):
    """
    Calculate phase velocities for various propagation directions.

    Parameters:
    - cijkl (array): The fourth-rank stiffness tensor for the material.
    - rho (float): Density of the material.

    Returns:
    - tuple: Tuples containing phase velocities for P-wave (vp), S-wave with higher velocity (vs1), and S-wave with lower velocity (vs2).
    """
    vp = []
    vs1 = []
    vs2 = []
    step = math.pi / 180

    for theta in np.arange(0, math.pi + step, step):
        for phi in np.arange(0, 2 * math.pi + step, step):
            #propagation direction vector n from theta and phi using spherical coordinates
            n = np.array([math.sin(theta) * math.cos(phi), math.sin(theta) * math.sin(phi), math.cos(theta)])
            # print(n)
            tik = christoffel_tensor(cijkl, n)
            wave_moduli, polarization_directions = wave_property(tik)
            vp.append(math.sqrt(wave_moduli[0] / rho))
            vs1.append(math.sqrt(wave_moduli[1] / rho))
            vs2.append(math.sqrt(wave_moduli[2] / rho))
    return tuple(vp), tuple(vs1), tuple(vs2)

def plot_vp_2d(cijkl, rho, save_plot=False, filename=None, dpi=300, cmap='RdBu'):
    """
    Plot phase velocities in a 2D stereographic projection.

    Parameters:
    - cijkl (array): The fourth-rank stiffness tensor for the material.
    - rho (float): Density of the material.
    - save_plot (bool): Whether to save the plot or not. Default is False.
    - filename (str): Name of the file to save the plot. Required if save_plot is True.
    - dpi (int): Dots per inch for saving the plot. Default is 300.
    - cmap (str or Colormap): The colormap to be used for the scatter plot. Default is 'RdBu'.

    Returns:
    - None
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)
    step = math.pi / 180
    x = []
    y = []
    c = []

    for theta in np.arange(0, math.pi / 2 + step, step):
        for phi in np.arange(0, 2 * math.pi + step, step):
            n = np.array([math.sin(theta) * math.cos(phi), math.sin(theta) * math.sin(phi), math.cos(theta)])
            tik = christoffel_tensor(cijkl, n)
            wave_moduli, polarization_directions = wave_property(tik)
            vp = math.sqrt(wave_moduli[0] / rho)
            x.append(n[0] / (1 + n[2]))
            y.append(n[1] / (1 + n[2]))
            c.append(vp)

    sc = ax.scatter(x, y, c=c, cmap=cmap)
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_title('vp in 2d stereographic projection plots with velocity represented by colour')
    cb = plt.colorbar(sc)
    cb.set_label('vp')
    cb.ax.set_yticklabels(['{:.1f}'.format(v) for v in cb.get_ticks()])

    if save_plot:
        if filename is None:
            raise ValueError("Filename must be provided when saving the plot.")
        plt.savefig(filename, dpi=dpi)
    else:
        plt.show()





import numpy as np

def christoffel_tensor(cijkl, n):
    tik = np.zeros((3, 3))

    for i in range(3):
        for k in range(3):
            for j in range(3):
                for l in range(3):
                    tik[i][k] += cijkl[i][j][k][l] * n[j] * n[l]
                    
    return tik

def wave_property(tik):
    eigenvalues, eigenvectors = np.linalg.eig(tik)
    indices = np.argsort(eigenvalues)[::-1]
    wave_moduli = []
    polarization_directions = []

    for i in indices:
        wave_moduli.append(eigenvalues[i])
        polarization_directions.append(eigenvectors[:, i] / np.linalg.norm(eigenvectors[:, i]))
    return tuple(wave_moduli), tuple(polarization_directions)


import numpy as np
class Tensor:
    def __init__(self, data=None):

        """
        Initialize a Tensor object.

        Parameters:
        - data (array): The data to initialize the tensor with. If None, an empty tensor is created.

        Returns:
        - None
        """
        self.data = np.array(data) if data is not None else None

    def tensor_to_voigt(self, tensor):

        """
        Convert a tensor to its Voigt notation.

        Parameters:
        - tensor (array): The input tensor.

        Returns:
        - array-like: The tensor converted to Voigt notation.
        """
        voigt = [[0 for i in range(6)] for j in range(6)]
        mapping = {(1, 1): 1, (2, 2): 2, (3, 3): 3, (2, 3): 4, (3, 2): 4, (1, 3): 5, (3, 1): 5, (1, 2): 6, (2, 1): 6}

        for i in range(1, 4):
            for j in range(1, 4):
                for k in range(1, 4):
                    for l in range(1, 4):

                        I = mapping[(i, j)]
                        J = mapping[(k, l)]
                        voigt[I-1][J-1] = tensor[i-1][j-1][k-1][l-1]

        return np.array(voigt)


    def voigt_to_tensor(self, voigt):
        """
        Convert Voigt notation back to a tensor.

        Parameters:
        - voigt (array): The input Voigt notation.

        Returns:
        - array: The Voigt notation converted back to a tensor.
        """
        tensor = [[[[0 for i in range(3)] for j in range(3)] for k in range(3)] for l in range(3)]
        mapping = {1: (1, 1), 2: (2, 2), 3: (3, 3), 4: (2, 3), 5: (1, 3), 6: (1, 2)}
        
        for I in range(1, 7):
            for J in range(1, 7):
                i, j = mapping[I]
                k, l = mapping[J]
                tensor[i-1][j-1][k-1][l-1] = voigt[I-1][J-1]
                tensor[j-1][i-1][k-1][l-1] = voigt[I-1][J-1]
                tensor[i-1][j-1][l-1][k-1] = voigt[I-1][J-1]
                tensor[j-1][i-1][l-1][k-1] = voigt[I-1][J-1]
        return np.array(tensor)
    
    def euler_to_rotation(self, phi1, phi, phi2):

        """
        Convert Euler angles to a rotation matrix.

        Parameters:
        - phi1 (float): First Euler angle in degrees.
        - phi (float): Second Euler angle in degrees.
        - phi2 (float): Third Euler angle in degrees.

        Returns:
        - array: The rotation matrix corresponding to the Euler angles.
        """
        phi1 = np.radians(phi1)
        phi = np.radians(phi)
        phi2 = np.radians(phi2)
        Rz1 = np.array([[np.cos(phi1), -np.sin(phi1), 0], [np.sin(phi1), np.cos(phi1), 0], [0, 0, 1]])
        Rx = np.array([[1, 0, 0], [0, np.cos(phi), -np.sin(phi)], [0, np.sin(phi), np.cos(phi)]])
        Rz2 = np.array([[np.cos(phi2), -np.sin(phi2), 0], [np.sin(phi2), np.cos(phi2), 0], [0, 0, 1]])
        R = Rz1 @ Rx @ Rz2 

        return R

    def rotate_tensor(self, tensor, phi1, phi, phi2):
        """
        Rotate a tensor using Euler angles.

        Parameters:
        - tensor (array-like): The input tensor to be rotated.
        - phi1 (float): First Euler angle in degrees.
        - phi (float): Second Euler angle in degrees.
        - phi2 (float): Third Euler angle in degrees.

        Returns:
        - array: The rotated tensor.
        """
        R = self.euler_to_rotation(phi1, phi, phi2)
        tensor = np.array(tensor)
        R_tensor = np.einsum('im,jn,ko,lp,mnop->ijkl', R, R, R, R, tensor)

        return R_tensor




import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde

from orix.crystal_map import Phase
from orix.quaternion import Orientation, symmetry
from orix.vector import Miller, Vector3d
import plotly.graph_objects as go

import random

def odf(df, phase=1, crystal_symmetry='D2', random_val=True, miller=[1, 0, 0], hemisphere = 'both', 
        axes_labels=["Xs", "Ys"], alpha = 0.01, figure = None, vector_labels = None, reproject=False, show_hemisphere_label = None,
        grid = None, grid_resolution = None, return_figure = None):
    """
        Calculate the Orientation Distribution Function (ODF) for a given EBSD dataset and optionally visualize it.

        Parameters:
        - df: pandas DataFrame
            The DataFrame containing EBSD (Electron Backscatter Diffraction) data.
        - phase: int, optional
            The phase number for which ODF needs to be calculated (default is 1).
        - crystal_symmetry: str, optional
            The crystal symmetry of the material, e.g., 'D2' for cubic symmetry (default is 'D2').
        - random_val: bool, optional
            If True, randomly generate orientation values for incomplete data (default is True).
        - miller: list of int, optional
            The Miller indices representing the crystallographic plane to calculate ODF for (default is [1, 0, 0]).
        - projection: str, optional
            The type of projection to use for visualization (default is 'stereographic').
        - figure: matplotlib Figure or None, optional
            If provided, the ODF plot will be added to this Figure; otherwise, a new Figure will be created.
        - axes_labels: list of str or None, optional
            Labels for the X and Y axes of the ODF plot (default is None).
        - vector_labels: list of str or None, optional
            Labels for the vectors in the ODF plot (default is None, which results in auto-generated labels).
        - hemisphere: str, upper or lower or both, optional
            Specify the hemisphere for ODF calculation and visualization (default is None).
        - reproject: bool, optional
            If True, reproject data into a specified coordinate system (default is False).
        - show_hemisphere_label: bool or None, optional
            Specify whether to show the hemisphere label on the plot (default is None).
        - grid: bool or None, optional
            Specify whether to display a grid on the ODF plot (default is None).
        - grid_resolution: tuple of float or None, optional
            Resolution of the grid (default is None).
        - figure_kwargs: dict or None, optional
            Additional keyword arguments for configuring the matplotlib Figure (default is None).
        - reproject_scatter_kwargs: dict or None, optional
            Additional keyword arguments for configuring scatter plot during reprojection (default is None).
        - text_kwargs: dict or None, optional
            Additional keyword arguments for configuring text elements in the plot (default is None).
        - return_figure: bool, optional
            If True, return the matplotlib Figure object along with the ODF data (default is False).

        Returns:
        - odf_result: pandas DataFrame
            DataFrame containing the calculated Orientation Distribution Function (ODF) values.
        - figure: matplotlib Figure or None
            If return_figure is True, the Figure object containing the ODF plot; otherwise, None.
        """
    
    euler_df = df[df['Phase'] == phase][['Euler1', 'Euler2', 'Euler3']]

    euler_values = [tuple(x) for x in euler_df.to_numpy()]
    if random_val == True:
        random.shuffle(euler_values)
        euler_values = euler_values[:1250]
    else:
        euler_values = euler_values

    if crystal_symmetry in ["triclinic", "C1", "1", "-1", "1"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C1)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C1))
    elif crystal_symmetry in ["triclinic", "Ci", "-1", "-1", "1"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Ci)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.Ci))
    elif crystal_symmetry in ["monoclinic", "C2", "211", "2/m11", "211"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2))
    elif crystal_symmetry in ["monoclinic", "Cs", "m11", "2/m11", "211"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Cs)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.Cs))
    elif crystal_symmetry in ["monoclinic", "C2h", "12/m1", "12/m1", "121"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2h))
    elif crystal_symmetry in ["monoclinic", "C2", "121", "12/m1", "121"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2))
    elif crystal_symmetry in ["monoclinic", "Cs", "1m1", "12/m1", "121"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Cs)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.Cs))
    elif crystal_symmetry in ["monoclinic", "C2h", "12/m1", "12/m1", "121"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2h))
    elif crystal_symmetry in ["monoclinic", "C2", "112", "112/m", "112"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2))
    elif crystal_symmetry in ["monoclinic", "Cs", "11m", "112/m", "112"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Cs)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.Cs))
    elif crystal_symmetry in ["monoclinic", "C2h", "112/m", "112/m", "112"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2h))
    elif crystal_symmetry in ["orthorhombic", "D2", "222", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D2)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D2))
    elif crystal_symmetry in ["orthorhombic", "C2v", "2mm", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2v)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2v))
    elif crystal_symmetry in ["orthorhombic", "C2v", "m2m", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2v)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2v))
    elif crystal_symmetry in ["orthorhombic", "C2v", "mm2", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2v)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2v))
    elif crystal_symmetry in ["orthorhombic", "D2h", "mmm", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D2h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D2h))
    elif crystal_symmetry in ["trigonal", "C3", "3", "-3", "3"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C3))
    elif crystal_symmetry in ["trigonal", "C3i", "-3", "-3", "3"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3i)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C3i))
    elif crystal_symmetry in ["trigonal", "D3", "321", "-3m1", "321"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D3))
    elif crystal_symmetry in ["trigonal", "C3v", "3m1", "-3m1", "321"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3v)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C3v))
    elif crystal_symmetry in ["trigonal", "D3d", "-3m1", "-3m1", "321"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3d)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D3d))
    elif crystal_symmetry in ["trigonal", "D3", "312", "-31m", "312"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D3))
    elif crystal_symmetry in ["trigonal", "C3v", "31m", "-31m", "312"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3v)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C3v))
    elif crystal_symmetry in ["trigonal", "D3d", "-31m", "-31m", "312"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3d)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D3d))
    elif crystal_symmetry in ["tetragonal", "C4", "4", "4/m", "4"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C4)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C4))
    elif crystal_symmetry in ["tetragonal", "S4", "-4", "4/m", "4"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.S4)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.S4))
    elif crystal_symmetry in ["tetragonal", "C4h", "4/m", "4/m", "4"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C4h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C4h))
    elif crystal_symmetry in ["tetragonal", "D4", "422", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D4)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D4))
    elif crystal_symmetry in ["tetragonal", "C4v", "4mm", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C4v)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C4v))
    elif crystal_symmetry in ["tetragonal", "D2d", "-42m", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D2d)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D2d))
    elif crystal_symmetry in ["tetragonal", "D2d", "-4m2", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D2d)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D2d))
    elif crystal_symmetry in ["tetragonal", "D4h", "4/mmm", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D4h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D4h))
    elif crystal_symmetry in ["hexagonal", "C6", "6", "6/m", "6"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C6)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C6))
    elif crystal_symmetry in ["hexagonal", "C3h", "-6", "6/m", "6"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C3h))
    elif crystal_symmetry in ["hexagonal", "C6h", "6/m", "6/m", "6"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C6h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C6h))
    elif crystal_symmetry in ["hexagonal", "D6", "622", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D6)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D6))
    elif crystal_symmetry in ["hexagonal", "C6v", "6mm", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C6v)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C6v))
    elif crystal_symmetry in ["hexagonal", "D3h", "-62m", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D3h))
    elif crystal_symmetry in ["hexagonal", "D3h", "-6m2", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D3h))
    elif crystal_symmetry in ["hexagonal", "D6h", "6/mmm", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D6h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D6h))
    elif crystal_symmetry in ["cubic", "T", "23", "m-3", "23"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.T)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.T))
    elif crystal_symmetry in ["cubic", "Th", "m-3", "m-3", "23"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Th)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.Th))
    elif crystal_symmetry in ["cubic", "O", "432", "m-3m", "432"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.O)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.O))
    elif crystal_symmetry in ["cubic", "Td", "-43m", "m-3m", "432"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Td)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.Td))
    elif crystal_symmetry in ["cubic", "Oh", "m-3m", "m-3m", "432"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Oh)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.Oh))

    uvw_ = uvw_.symmetrise(unique=False)

    uvw_ = uvw_.reshape(1, uvw_.size)
    orientations = orientations.reshape(*orientations.shape, 1)
    r_ = ~orientations * uvw_

    r_.scatter(hemisphere=hemisphere, axes_labels=axes_labels, alpha = alpha, figure = figure, vector_labels = vector_labels, reproject=reproject, show_hemisphere_label = show_hemisphere_label,
        grid = grid, grid_resolution = grid_resolution, return_figure = return_figure)
    return r_

def ipf(df, phase=1, vector_sample=[0, 0, 1], random_val=True,
            vector_title='Z', projection='ipf', crystal_symmetry='D2'):
    """
    Generate and visualize an Inverse Pole Figure (IPF) from EBSD data.

    Parameters:
    - df: pandas DataFrame
        The DataFrame containing EBSD (Electron Backscatter Diffraction) data.
    - phase: int, optional
        The phase number for which IPF needs to be generated (default is 1).
    - vector_sample: list of int, optional
        The sample vector used for IPF generation (default is [0, 0, 1]).
    - random_val: bool, optional
        If True, randomly select orientation values for IPF generation (default is True).
    - vector_title: str, optional
        The title for the vector used in IPF visualization (default is 'Z').
    - projection: str, optional
        The type of projection for IPF visualization (default is 'ipf').
    - crystal_symmetry: str, optional
        The crystal symmetry of the material (default is 'D2').

    Returns:
    - fig: matplotlib Figure
        The generated matplotlib Figure object containing the IPF visualization.
    """
    
    from orix.quaternion import Orientation, symmetry
    from orix.vector import Miller, Vector3d
    import matplotlib.pyplot as plt
    from orix.crystal_map import Phase

    euler_df = df[df['Phase'] == phase][['Euler1', 'Euler2', 'Euler3']]

    euler_values = [tuple(x) for x in euler_df.to_numpy()]
    if random_val == True:
        random.shuffle(euler_values)
        euler_values = euler_values[:1250]
    else:
        euler_values = euler_values

    if crystal_symmetry in ["triclinic", "C1", "1", "-1", "1"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C1)
        symmetry_ = symmetry.C1
    elif crystal_symmetry in ["triclinic", "Ci", "-1", "-1", "1"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Ci)
        symmetry_ = symmetry.Ci
    elif crystal_symmetry in ["monoclinic", "C2", "211", "2/m11", "211"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2)
        symmetry_ = symmetry.C2
    elif crystal_symmetry in ["monoclinic", "Cs", "m11", "2/m11", "211"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Cs)
        symmetry_ = symmetry.Cs
    elif crystal_symmetry in ["monoclinic", "C2h", "12/m1", "12/m1", "121"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2h)
        symmetry_ =symmetry.C2h
    elif crystal_symmetry in ["monoclinic", "C2", "121", "12/m1", "121"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2)
        symmetry_ =symmetry.C2
    elif crystal_symmetry in ["monoclinic", "Cs", "1m1", "12/m1", "121"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Cs)
        symmetry_ =symmetry.Cs
    elif crystal_symmetry in ["monoclinic", "C2h", "12/m1", "12/m1", "121"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2h)
        symmetry_ =symmetry.C2h
    elif crystal_symmetry in ["monoclinic", "C2", "112", "112/m", "112"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2)
        symmetry_ =symmetry.C2
    elif crystal_symmetry in ["monoclinic", "Cs", "11m", "112/m", "112"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Cs)
        symmetry_ =symmetry.Cs
    elif crystal_symmetry in ["monoclinic", "C2h", "112/m", "112/m", "112"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2h)
        symmetry_ =symmetry.C2h
    elif crystal_symmetry in ["orthorhombic", "D2", "222", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D2)
        symmetry_ =symmetry.D2
    elif crystal_symmetry in ["orthorhombic", "C2v", "2mm", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2v)
        symmetry_ =symmetry.C2v
    elif crystal_symmetry in ["orthorhombic", "C2v", "m2m", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2v)
        symmetry_ =symmetry.C2v
    elif crystal_symmetry in ["orthorhombic", "C2v", "mm2", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2v)
        symmetry_ =symmetry.C2v
    elif crystal_symmetry in ["orthorhombic", "D2h", "mmm", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D2h)
        symmetry_ =symmetry.D2h
    elif crystal_symmetry in ["trigonal", "C3", "3", "-3", "3"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3)
        symmetry_ =symmetry.C3
    elif crystal_symmetry in ["trigonal", "C3i", "-3", "-3", "3"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3i)
        symmetry_ =symmetry.C3i
    elif crystal_symmetry in ["trigonal", "D3", "321", "-3m1", "321"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3)
        symmetry_ =symmetry.D3
    elif crystal_symmetry in ["trigonal", "C3v", "3m1", "-3m1", "321"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3v)
        symmetry_ =symmetry.C3v
    elif crystal_symmetry in ["trigonal", "D3d", "-3m1", "-3m1", "321"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3d)
        symmetry_ =symmetry.D3d
    elif crystal_symmetry in ["trigonal", "D3", "312", "-31m", "312"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3)
        symmetry_ =symmetry.D3
    elif crystal_symmetry in ["trigonal", "C3v", "31m", "-31m", "312"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3v)
        symmetry_ =symmetry.C3v
    elif crystal_symmetry in ["trigonal", "D3d", "-31m", "-31m", "312"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3d)
        symmetry_ =symmetry.D3d
    elif crystal_symmetry in ["tetragonal", "C4", "4", "4/m", "4"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C4)
        symmetry_ =symmetry.C4
    elif crystal_symmetry in ["tetragonal", "S4", "-4", "4/m", "4"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.S4)
        symmetry=symmetry.S4
    elif crystal_symmetry in ["tetragonal", "C4h", "4/m", "4/m", "4"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C4h)
        symmetry_ =symmetry.C4h
    elif crystal_symmetry in ["tetragonal", "D4", "422", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D4)
        symmetry_ =symmetry.D4
    elif crystal_symmetry in ["tetragonal", "C4v", "4mm", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C4v)
        symmetry_ =symmetry.C4v
    elif crystal_symmetry in ["tetragonal", "D2d", "-42m", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D2d)
        symmetry_ =symmetry.D2d
    elif crystal_symmetry in ["tetragonal", "D2d", "-4m2", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D2d)
        symmetry_ =symmetry.D2d
    elif crystal_symmetry in ["tetragonal", "D4h", "4/mmm", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D4h)
        symmetry_ =symmetry.D4h
    elif crystal_symmetry in ["hexagonal", "C6", "6", "6/m", "6"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C6)
        symmetry_ =symmetry.C6
    elif crystal_symmetry in ["hexagonal", "C3h", "-6", "6/m", "6"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3h)
        symmetry_ =symmetry.C3h
    elif crystal_symmetry in ["hexagonal", "C6h", "6/m", "6/m", "6"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C6h)
        symmetry_ =symmetry.C6h
    elif crystal_symmetry in ["hexagonal", "D6", "622", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D6)
        symmetry_ =symmetry.D6
    elif crystal_symmetry in ["hexagonal", "C6v", "6mm", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C6v)
        symmetry_ =symmetry.C6v
    elif crystal_symmetry in ["hexagonal", "D3h", "-62m", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3h)
        symmetry_ =symmetry.D3h
    elif crystal_symmetry in ["hexagonal", "D3h", "-6m2", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3h)
        symmetry_ =symmetry.D3h
    elif crystal_symmetry in ["hexagonal", "D6h", "6/mmm", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D6h)
        symmetry_ =symmetry.D6h
    elif crystal_symmetry in ["cubic", "T", "23", "m-3", "23"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.T)
        symmetry_ =symmetry.T
    elif crystal_symmetry in ["cubic", "Th", "m-3", "m-3", "23"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Th)
        symmetry_ =symmetry.Th
    elif crystal_symmetry in ["cubic", "O", "432", "m-3m", "432"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.O)
        symmetry_ =symmetry.O
    elif crystal_symmetry in ["cubic", "Td", "-43m", "m-3m", "432"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Td)
        symmetry_ =symmetry.Td
    elif crystal_symmetry in ["cubic", "Oh", "m-3m", "m-3m", "432"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Oh)
        symmetry_ =symmetry.Oh

    orientations = orientations.reshape(*orientations.shape, 1)
    # print(orientations)
    vec_sample = Vector3d(vector_sample)
    # print(vec_sample)
    vec_crystal = orientations * vec_sample
    subplot_kw = dict(projection=projection, symmetry=symmetry_)
    fig = plt.figure(figsize=(9, 8))

    ax0 = fig.add_subplot(221, direction=vec_sample, **subplot_kw)
    ax0.scatter(orientations, alpha=0.05)
    _ = ax0.set_title(f"Phase{phase}, {vector_title}")

    ax2 = fig.add_subplot(223, direction=vec_sample, **subplot_kw)
    ax2.pole_density_function(vec_crystal)
    _ = ax2.set_title(f"Phase{phase}, {vector_title}")


def pdf(df, phase=1, crystal_symmetry='D2', random_val=True, miller=[0, 1, 0], hemisphere = 'both', sigma = 4,
        axes_labels=["Xs", "Ys"], figure = None, show_hemisphere_label = None,
        grid = None, grid_resolution = None, return_figure = None, log = False, colorbar=True, weights = None):
    
    """
        Calculate the Orientation Distribution Function (ODF) for a given EBSD dataset and optionally visualize it.

        Parameters:
        - df: pandas DataFrame
            The DataFrame containing EBSD (Electron Backscatter Diffraction) data.
        - phase: int, optional
            The phase number for which ODF needs to be calculated (default is 1).
        - crystal_symmetry: str, optional
            The crystal symmetry of the material, e.g., 'D2' for cubic symmetry (default is 'D2').
        - random_val: bool, optional
            If True, randomly generate orientation values for incomplete data (default is True).
        - miller: list of int, optional
            The Miller indices representing the crystallographic plane to calculate ODF for (default is [1, 0, 0]).
        - axes_labels: list of str or None, optional
            Labels for the X and Y axes of the ODF plot (default is None).
        - hemisphere: str, upper or lower or both, optional
            Specify the hemisphere for ODF calculation and visualization (default is None).
        - reproject: bool, optional
            If True, reproject data into a specified coordinate system (default is False).
        - show_hemisphere_label: bool or None, optional
            Specify whether to show the hemisphere label on the plot (default is None).
        - grid: bool or None, optional
            Specify whether to display a grid on the ODF plot (default is None).
        - grid_resolution: tuple of float or None, optional
            Resolution of the grid (default is None).
        - figure_kwargs: dict or None, optional
            Additional keyword arguments for configuring the matplotlib Figure (default is None).
        - reproject_scatter_kwargs: dict or None, optional
            Additional keyword arguments for configuring scatter plot during reprojection (default is None).
        - text_kwargs: dict or None, optional
            Additional keyword arguments for configuring text elements in the plot (default is None).
        - return_figure: bool, optional
            If True, return the matplotlib Figure object along with the ODF data (default is False).

        Returns:
        - odf_result: pandas DataFrame
            DataFrame containing the calculated Orientation Distribution Function (ODF) values.
        - figure: matplotlib Figure or None
            If return_figure is True, the Figure object containing the ODF plot; otherwise, None.
        """
    
    euler_df = df[df['Phase'] == phase][['Euler1', 'Euler2', 'Euler3']]

    euler_values = [tuple(x) for x in euler_df.to_numpy()]
    if random_val == True:
        random.shuffle(euler_values)
        euler_values = euler_values[:1250]
    else:
        euler_values = euler_values

    if crystal_symmetry in ["triclinic", "C1", "1", "-1", "1"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C1)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C1))
    elif crystal_symmetry in ["triclinic", "Ci", "-1", "-1", "1"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Ci)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.Ci))
    elif crystal_symmetry in ["monoclinic", "C2", "211", "2/m11", "211"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2))
    elif crystal_symmetry in ["monoclinic", "Cs", "m11", "2/m11", "211"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Cs)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.Cs))
    elif crystal_symmetry in ["monoclinic", "C2h", "12/m1", "12/m1", "121"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2h))
    elif crystal_symmetry in ["monoclinic", "C2", "121", "12/m1", "121"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2))
    elif crystal_symmetry in ["monoclinic", "Cs", "1m1", "12/m1", "121"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Cs)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.Cs))
    elif crystal_symmetry in ["monoclinic", "C2h", "12/m1", "12/m1", "121"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2h))
    elif crystal_symmetry in ["monoclinic", "C2", "112", "112/m", "112"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2))
    elif crystal_symmetry in ["monoclinic", "Cs", "11m", "112/m", "112"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Cs)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.Cs))
    elif crystal_symmetry in ["monoclinic", "C2h", "112/m", "112/m", "112"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2h))
    elif crystal_symmetry in ["orthorhombic", "D2", "222", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D2)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D2))
    elif crystal_symmetry in ["orthorhombic", "C2v", "2mm", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2v)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2v))
    elif crystal_symmetry in ["orthorhombic", "C2v", "m2m", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2v)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2v))
    elif crystal_symmetry in ["orthorhombic", "C2v", "mm2", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C2v)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C2v))
    elif crystal_symmetry in ["orthorhombic", "D2h", "mmm", "mmm", "222"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D2h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D2h))
    elif crystal_symmetry in ["trigonal", "C3", "3", "-3", "3"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C3))
    elif crystal_symmetry in ["trigonal", "C3i", "-3", "-3", "3"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3i)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C3i))
    elif crystal_symmetry in ["trigonal", "D3", "321", "-3m1", "321"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D3))
    elif crystal_symmetry in ["trigonal", "C3v", "3m1", "-3m1", "321"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3v)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C3v))
    elif crystal_symmetry in ["trigonal", "D3d", "-3m1", "-3m1", "321"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3d)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D3d))
    elif crystal_symmetry in ["trigonal", "D3", "312", "-31m", "312"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D3))
    elif crystal_symmetry in ["trigonal", "C3v", "31m", "-31m", "312"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3v)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C3v))
    elif crystal_symmetry in ["trigonal", "D3d", "-31m", "-31m", "312"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3d)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D3d))
    elif crystal_symmetry in ["tetragonal", "C4", "4", "4/m", "4"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C4)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C4))
    elif crystal_symmetry in ["tetragonal", "S4", "-4", "4/m", "4"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.S4)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.S4))
    elif crystal_symmetry in ["tetragonal", "C4h", "4/m", "4/m", "4"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C4h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C4h))
    elif crystal_symmetry in ["tetragonal", "D4", "422", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D4)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D4))
    elif crystal_symmetry in ["tetragonal", "C4v", "4mm", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C4v)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C4v))
    elif crystal_symmetry in ["tetragonal", "D2d", "-42m", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D2d)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D2d))
    elif crystal_symmetry in ["tetragonal", "D2d", "-4m2", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D2d)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D2d))
    elif crystal_symmetry in ["tetragonal", "D4h", "4/mmm", "4/mmm", "422"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D4h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D4h))
    elif crystal_symmetry in ["hexagonal", "C6", "6", "6/m", "6"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C6)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C6))
    elif crystal_symmetry in ["hexagonal", "C3h", "-6", "6/m", "6"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C3h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C3h))
    elif crystal_symmetry in ["hexagonal", "C6h", "6/m", "6/m", "6"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C6h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C6h))
    elif crystal_symmetry in ["hexagonal", "D6", "622", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D6)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D6))
    elif crystal_symmetry in ["hexagonal", "C6v", "6mm", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.C6v)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.C6v))
    elif crystal_symmetry in ["hexagonal", "D3h", "-62m", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D3h))
    elif crystal_symmetry in ["hexagonal", "D3h", "-6m2", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D3h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D3h))
    elif crystal_symmetry in ["hexagonal", "D6h", "6/mmm", "6/mmm", "622"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.D6h)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.D6h))
    elif crystal_symmetry in ["cubic", "T", "23", "m-3", "23"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.T)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.T))
    elif crystal_symmetry in ["cubic", "Th", "m-3", "m-3", "23"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Th)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.Th))
    elif crystal_symmetry in ["cubic", "O", "432", "m-3m", "432"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.O)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.O))
    elif crystal_symmetry in ["cubic", "Td", "-43m", "m-3m", "432"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Td)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.Td))
    elif crystal_symmetry in ["cubic", "Oh", "m-3m", "m-3m", "432"]:
        orientations = Orientation.from_euler(np.deg2rad(euler_values), symmetry.Oh)
        uvw_ = Miller(uvw = miller, phase=Phase(point_group=symmetry.Oh))

    uvw_ = uvw_.symmetrise(unique=False)

    uvw_ = uvw_.reshape(1, uvw_.size)
    orientations = orientations.reshape(*orientations.shape, 1)
    r_ = ~orientations * uvw_
    r_.pole_density_function(sigma=sigma, hemisphere=hemisphere, axes_labels=axes_labels, figure = figure, show_hemisphere_label = show_hemisphere_label,
        grid = grid, grid_resolution = grid_resolution, return_figure = return_figure, log = log, colorbar=colorbar, weights = weights)


def calculate_odf(euler_angles):
    """
    Calculate and plot an ODF from Bunge Euler angles.
    
    Parameters:
    - euler_angles: A Nx3 numpy array of Euler angles (phi1, Phi, phi2) in degrees.
    
    """
    euler_angles_rad = np.radians(euler_angles)
    
    # using a Gaussian KDE to estimate the distribution
    kde = gaussian_kde(euler_angles_rad.T)
    
    phi1, Phi, phi2 = np.mgrid[0:np.pi:100j, 0:np.pi/2:100j, 0:2*np.pi:100j]
    positions = np.vstack([phi1.ravel(), Phi.ravel(), phi2.ravel()])
    
    density = np.reshape(kde(positions).T, phi1.shape)
    
    fig = plt.figure(figsize=(10, 8))
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(phi1, Phi, phi2, c=density, cmap='viridis')
    ax.set_xlabel('Phi1')
    ax.set_ylabel('Phi')
    ax.set_zlabel('Phi2')
    ax.set_title('ODF Representation')
    plt.show()


def plot_orientations_3d(euler_angles):

    phi1, Phi, phi2 = euler_angles[:, 0], euler_angles[:, 1], euler_angles[:, 2]
    
    fig = go.Figure(data=[go.Scatter3d(
        x=phi1,
        y=Phi,
        z=phi2,
        mode='markers',
        marker=dict(
            size=5,
            color='rgba(255, 0, 0, 0.8)',  
        )
    )])
    
    fig.update_layout(
        title='3D Plot of Bunge-Euler Orientations',
        scene=dict(
            xaxis_title='Phi1 (Degrees)',
            yaxis_title='Phi (Degrees)',
            zaxis_title='Phi2 (Degrees)'
        ),
        margin=dict(r=0, b=0, l=0, t=30) 
    )
    
    fig.show()




from setuptools import setup, find_packages

VERSION = '1.1'
DESCRIPTION = "SAnTeX is a Python library which calculates seismic anisotropy from full elastic tensor of rocks from modal mineral composition, crystallographic orientation, and a crystal stiffness tensor catalogue that accounts for the dependency of elasticity with pressure and temperature. SAnTex facilitates the processing and cleaning of EBSD data and calculation of Orientation Distribution Function (ODF) and Inverse pole figure (IPF)"

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='santex',
    version=VERSION,
    author='Utpal Singh',
    author_email="utpal_singh@yahoo.com",
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/utpal-singh/SAnTex",
    license='GNU Lesser General Public License v3 (LGPLv3)',
    classifiers=[
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: OS Independent',
    ],
    packages=find_packages(),
    install_requires=[
        'joblib',
        'pandas',
        'matplotlib',
        'numpy',
        'vtk',
        'tqdm',
        'tabulate',
        'scikit-learn',
        'plotly',
        'orix',
        'sphinx_rtd_theme',
    ],
    keywords=['python', 'seismic', 'seismic anisotropy', 'EBSD', 'Texture', "Crystallography"],
    package_data={
        'santex': [
            'isotropy/data/*',
            'material/data/*',
        ],
    },
    include_package_data=True,
)



import numpy as np
from ..tensor.tensor import Tensor


def calcMelttensor():
    """
    Returns the melt stiffness tensor.
    
    Parameters: 
    - None
    
    Returns:
    - np.ndarray - melt stifness tensor
    
    Reference 
    Lee, A. L., Walker, A. M., Lloyd, G. E., & Torvela, T. (2017). 
    Modeling the impact of melt on seismic properties during mountain building. 
    Geochemistry, Geophysics, Geosystems, 18(3), 1090–1110. https://doi.org/10.1002/2016GC006705
    """
    voigt_melt = np.array([[16.1,   15.967,  15.967,   0.,      0,   0.   ],
                                [ 15.967, 16.1,   15.967,    0.,      0,   0.   ],
                                [ 15.967,  15.967,  16.1,    0.,     0,    0.   ],
                                [  0.,      0.,      0.,     0.01,    0.,      0],
                                [  0,   0,  0,    0.,     0.01,    0.   ],
                                [  0.,      0.,     0.,      0,   0.,     0.01 ]])



    tensor = Tensor()

    tensor_melt = tensor.voigt_to_tensor(voigt_melt)
    return tensor_melt


import numpy as np
import matplotlib.pyplot as plt

def plot_velocity_grid(pressure_range, temperature_range, return_type, is_ebsd = False, phase = None, grid = [5, 5], filename = None,fig_name=None, save_plot=False, dpi=300, *args):
    from santex import Anisotropy
    from santex import EBSD
    from santex import Material
    """
    Generates a grid of velocity values based on specified pressure and temperature ranges.

    Parameters:
        pressure_range (tuple): A tuple containing the minimum and maximum pressure values (in GPa).
        temperature_range (tuple): A tuple containing the minimum and maximum temperature values (in Kelvin).
        return_type (str): The type of velocity values to return. Allowed values are 'maxvp', 'minv', 'maxvs1', 'minvs1', 'maxvs2', 'minvs2'.
        is_ebsd (bool, optional): Indicates whether EBSD data is used. Default is False.
        phase (str, optional): The phase for which velocity values are calculated. Required if is_ebsd is False.
        grid (list, optional): A list specifying the grid size as [rows, columns]. Default is [5, 5].
        filename (str, optional): The filename of the EBSD data file. Required if is_ebsd is True.
        *args: Additional phase indices if EBSD data is used.

    Returns:
        np.ndarray: A grid of velocity values corresponding to the specified pressure and temperature ranges.

    Notes:
        - If is_ebsd is False, phase must be provided to calculate velocity values for the specified phase.
        - If is_ebsd is True, filename must be provided along with additional phase indices (*args) to calculate velocity values from EBSD data.
        - The return_type parameter determines the type of velocity values returned, such as maximum velocity (maxvp), minimum velocity (minv), etc.
    
    Usage:
    args can be [0, 1, 2, 3]
    give filename is is_ebsd is True
    """

    if is_ebsd is False:
        cij = []
        density = []

        pressure_grid = np.linspace(pressure_range[0], pressure_range[1], grid[0])
        temperature_grid = np.linspace(temperature_range[0], temperature_range[1], grid[1])
        vp_matrix = np.zeros((grid[0], grid[1]))
        
        for i, p in enumerate(pressure_grid):
            for j, t in enumerate(temperature_grid):
                material = Material()
                cij_highpt = material.voigthighPT(phase, p, t)*10**9
                density_highpt = material.load_density(phase, p, t)
                anisotropy = Anisotropy(cij_highpt, density_highpt)
                vp_matrix[i, j] = anisotropy.anisotropy_values(return_values=return_type)
        pressure_mesh, temperature_mesh = np.meshgrid(pressure_grid, temperature_grid)
        plt.figure(figsize=(8, 6))
        plt.pcolormesh(pressure_mesh, temperature_mesh, vp_matrix, shading='auto', cmap='viridis')
        plt.colorbar(label=f'{return_type}')
        plt.xlabel('Pressure')
        plt.ylabel('Temperature')
        plt.title(f'{return_type} vs. Pressure and Temperature')
        # plt.grid(True)
        plt.tight_layout()
        if save_plot:
            if fig_name is None:
                raise ValueError("Filename must be provided when saving the plot.")
            plt.savefig(fig_name, dpi=dpi)
        else:
            plt.show()

        return vp_matrix

    if is_ebsd is True:
        euler_angles = []
        df = ebsd.get_ebsd_data()
        for i in args:
            ebsd = EBSD(filename)
            euler_angles.append(ebsd.get_euler_angles(phase=i, data = df))

        average_tensor, average_density = ebsd.getAnisotropyForEBSD(cij, euler_angles, density)
        anis = Anisotropy(average_tensor*10**9, average_density)
    



